# Copyright 2021 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""CiscoNX generator."""

from typing import Optional

from aerleon.lib import aclgenerator, cisco


class Error(aclgenerator.Error):
    """Base error class."""


class UnsupportedNXosAccessListError(Error):
    """When a filter type is not supported in an NXOS policy target."""


class CiscoNX(cisco.Cisco):
    """An CiscoNX policy object.

    CiscoNX devices differ slightly from Cisco, omitting the extended argument to
    ACLs for example.
    """

    _PLATFORM = 'cisconx'
    SUFFIX = '.nxacl'
    # Protocols should be emitted as they were in the policy (names).
    _PROTO_INT = False

    def _RepositoryTagsHelper(
        self, target: Optional[list[str]] = None, filter_type: str = '', filter_name: str = ''
    ) -> list[str]:
        if target is None:
            target = []
        target.extend(aclgenerator.AddRepositoryTags(' remark ', rid=False, wrap=True))
        return target

    # CiscoNX omits the "extended" access-list argument.
    def _AppendTargetByFilterType(self, filter_name: str, filter_type: str) -> list[str]:
        """Takes in the filter name and type and appends headers.

        Args:
          filter_name: Name of the current filter
          filter_type: Type of current filter

        Returns:
          list of strings

        Raises:
          UnsupportedNXosAccessListError: When unknown filter type is used.
        """
        target = []
        if filter_type == 'extended':
            target.append(f'no ip access-list {filter_name}')
            target.append(f'ip access-list {filter_name}')
        elif filter_type == 'object-group':
            target.append(f'no ip access-list {filter_name}')
            target.append(f'ip access-list {filter_name}')
        elif filter_type == 'inet6':
            target.append(f'no ipv6 access-list {filter_name}')
            target.append(f'ipv6 access-list {filter_name}')
        else:
            raise UnsupportedNXosAccessListError(
                f'access list type {filter_type} not supported by {self._PLATFORM}'
            )
        return target



# Copyright 2018-2021 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
"""Google Cloud Armor Firewall Generator.

Refer to the links below for more information
https://cloud.google.com/armor/
https://cloud.google.com/armor/docs/
"""

import copy
import json
from typing import TypedDict

from absl import logging

from aerleon.lib import aclgenerator, policy


# Generic error class
class Error(aclgenerator.Error):
    """Generic error class."""


class ExceededMaxTermsError(Error):
    """Raised when number of terms in a policy exceed _MAX_RULES_PER_POLICY."""


class UnsupportedFilterTypeError(Error):
    """Raised when unsupported filter type (i.e address family) is specified."""


class RuleMatchConfig(TypedDict):
    srcIpRanges: 'list[str]'


class RuleMatch(TypedDict):
    config: RuleMatchConfig
    versionedExpr: str


class PolicyRule(TypedDict):
    action: str
    description: str
    match: RuleMatch
    preview: bool
    priority: int


class Term(aclgenerator.Term):
    """Generates the Term for CloudArmor."""

    # Max srcIpRanges within a single term
    _MAX_IP_RANGES_PER_TERM = 5

    ACTION_MAP = {'accept': 'allow', 'deny': 'deny(404)'}

    _MAX_TERM_COMMENT_LENGTH = 64

    def __init__(
        self, term: policy.Term, address_family: str = 'inet', verbose: bool = True
    ) -> None:
        super().__init__(term)
        self.term = term
        self.address_family = address_family
        self.verbose = verbose

    def __str__(self) -> str:
        return ''

    def ConvertToDict(self, priority_index: int) -> list[PolicyRule]:
        """Converts term to dictionary representation of CloudArmor's JSON format.

        Takes all of the attributes associated with a term (match, action, etc) and
        converts them into a dictionary which most closely represents
        the CloudArmor API's JSON rule format. Additionally, splits a single term
        into multiple terms if the number of srcIpRanges exceed
        _MAX_IP_RANGES_PER_TERM.

        Args:
          priority_index: An integer priority value assigned to the term. In case
            the term is split into i sub-terms, the ith sub-term has
            priority = priority_index + i

        Returns:
          A list of dicts where each dict is a term

        Raises:
          UnsupportedFilterTypeError: Raised when an unsupported filter type is
            specified
        """
        term_dict = {}
        rules = []

        if self.term.comment and self.verbose:
            raw_comment = ' '.join(self.term.comment)
            if len(raw_comment) > self._MAX_TERM_COMMENT_LENGTH:
                term_dict['description'] = raw_comment[: self._MAX_TERM_COMMENT_LENGTH]
                logging.warning(
                    'Term comment exceeds maximum length = %d; Truncating ' 'comment..',
                    self._MAX_TERM_COMMENT_LENGTH,
                )
            else:
                term_dict['description'] = raw_comment

        term_dict['action'] = self.ACTION_MAP[self.term.action[0]]
        term_dict['preview'] = False

        if self.address_family == 'inet':
            saddrs = self.term.GetAddressOfVersion('source_address', 4)
        elif self.address_family == 'inet6':
            saddrs = self.term.GetAddressOfVersion('source_address', 6)
        elif self.address_family == 'mixed':
            saddrs = self.term.GetAddressOfVersion(
                'source_address', 4
            ) + self.term.GetAddressOfVersion('source_address', 6)
        else:
            raise UnsupportedFilterTypeError(f"'{self.address_family}' is not a valid filter type")

        term_dict['match'] = {
            'versionedExpr': 'SRC_IPS_V1',
            'config': {
                'srcIpRanges': saddrs,
            },
        }
        # If scrIpRanges within a single term exceed _MAX_IP_RANGES_PER_TERM,
        # split into multiple terms
        source_addr_chunks = [
            saddrs[x : x + self._MAX_IP_RANGES_PER_TERM]
            for x in range(0, len(saddrs), self._MAX_IP_RANGES_PER_TERM)
        ]

        if not source_addr_chunks:
            rule = copy.deepcopy(term_dict)
            rule['priority'] = priority_index
            rule['match']['config']['srcIpRanges'] = ['*']
            rules.append(rule)

        else:
            split_rule_count = len(source_addr_chunks)
            for i, chunk in enumerate(source_addr_chunks):
                rule = copy.deepcopy(term_dict)
                if split_rule_count > 1:
                    term_position_suffix = ' [%d/%d]' % (i + 1, split_rule_count)
                    desc_limit = self._MAX_TERM_COMMENT_LENGTH - len(term_position_suffix)
                    rule['description'] = (
                        rule.get('description', '')[:desc_limit] + term_position_suffix
                    )

                rule['priority'] = priority_index + i
                rule['match'] = {
                    'versionedExpr': 'SRC_IPS_V1',
                    'config': {
                        'srcIpRanges': [str(saddr) for saddr in chunk],
                    },
                }
                rules.append(rule)

        if len(source_addr_chunks) > 1:
            logging.debug(
                f'Current term {self.term.name} was split into {len(source_addr_chunks)} sub-terms since '
                '_MAX_IP_RANGES_PER_TERM was exceeded',
            )
        return rules


class CloudArmor(aclgenerator.ACLGenerator):
    """A CloudArmor policy object."""

    _PLATFORM = 'cloudarmor'
    SUFFIX = '.gca'
    _SUPPORTED_AF = {'inet', 'inet6', 'mixed'}

    # Maximum number of rules that a CloudArmor policy can contain
    _MAX_RULES_PER_POLICY = 200

    # Warn user when rule count exceeds this number
    _RULECOUNT_WARN_THRESHOLD = 190

    # Maps indiviudal filter options to their index positions in the POL header
    _FILTER_OPTIONS_MAP = {'filter_type': 0}

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, _ = super()._BuildTokens()
        supported_tokens -= {
            'destination_address',
            'destination_address_exclude',
            'destination_port',
            'expiration',
            'icmp_type',
            'stateless_reply',
            'option',
            'protocol',
            'platform',
            'platform_exclude',
            'source_address_exclude',
            'source_port',
            'verbatim',
        }
        supported_sub_tokens = {'action': {'accept', 'deny'}}
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        """Translates a Aerleon policy into a CloudArmor-specific data structure.

        Takes in a POL file, parses each term and populates the cloudarmor_policies
        list. Each term in this list is a dictionary formatted according to
        CloudArmor's rule API specification.

        Args:
          pol: A Policy() object representing a given POL file.
          exp_info: An int that specifies number of weeks till policy expiry.

        Raises:
          ExceededMaxTermsError: Raised when the number of terms in a policy exceed
            _MAX_RULES_PER_POLICY.

          UnsupportedFilterTypeError: Raised when an unsupported filter type is
            specified
        """
        self.cloudarmor_policies = []

        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)
            verbose = True
            if 'noverbose' in filter_options:
                filter_options.remove('noverbose')
                verbose = False

            if filter_options is None or not filter_options:
                filter_type = 'inet'
                logging.warning('No filter_type specified. Defaulting to inet (IPv4)')

            else:
                filter_type = filter_options[self._FILTER_OPTIONS_MAP['filter_type']]
                if filter_type not in self._SUPPORTED_AF:
                    raise UnsupportedFilterTypeError(f"'{filter_type}' is not a valid filter type")

            counter = 1

            for term in terms:
                json_rule_list = Term(
                    term, address_family=filter_type, verbose=verbose
                ).ConvertToDict(priority_index=counter)
                # count number of rules generated after split (if any)
                split_rule_count = len(json_rule_list)

                self.cloudarmor_policies.extend(json_rule_list)

                counter = counter + split_rule_count

                total_rule_count = len(self.cloudarmor_policies)

                if total_rule_count > self._RULECOUNT_WARN_THRESHOLD:
                    if total_rule_count > self._MAX_RULES_PER_POLICY:
                        raise ExceededMaxTermsError(
                            'Exceeded maximum number of rules in '
                            ' a single policy | MAX = %d' % self._MAX_RULES_PER_POLICY
                        )
                    else:
                        logging.warning(
                            'Current rule count (%d) is almost at maximum ' 'limit of %d',
                            total_rule_count,
                            self._MAX_RULES_PER_POLICY,
                        )

    def __str__(self) -> str:
        """Return the JSON blob for CloudArmor."""

        out = '%s\n\n' % (
            json.dumps(self.cloudarmor_policies, indent=2, separators=(',', ': '), sort_keys=True)
        )
        return out



# Copyright 2020-2021 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
"""Google Cloud Hierarchical Firewall Generator.

Hierarchical Firewalls (HF) are represented in a SecurityPolicy GCP resouce.
"""

import copy
import re
from typing import TypedDict, Union

from absl import logging

from aerleon.lib import gcp, nacaddr, policy


class ExceededCostError(gcp.Error):
    """Raised when the total cost of a policy is above the maximum."""


class DifferentPolicyNameError(gcp.Error):
    """Raised when headers in the same policy have a different policy name."""


class ApiVersionSyntaxMap:
    """Defines the syntax changes between different API versions.

    https://cloud.google.com/compute/docs/reference/rest/v1/firewallPolicies/addRule
    https://cloud.google.com/compute/docs/reference/rest/beta/organizationSecurityPolicies/addRule
    """

    SYNTAX_MAP = {
        'beta': {
            'display_name': 'displayName',
            'dest_ip_range': 'destIpRanges',
            'src_ip_range': 'srcIpRanges',
            'layer_4_config': 'layer4Configs',
        },
        'ga': {
            'display_name': 'shortName',
            'dest_ip_range': 'destIpRanges',
            'src_ip_range': 'srcIpRanges',
            'layer_4_config': 'layer4Configs',
        },
    }


class L4Matcher(TypedDict):
    IPProtocol: str
    ports: 'list[str]'


class MatchConfig(TypedDict):
    destIpRanges: 'list[str]'
    srcIpRanges: 'list[str]'
    layer4Configs: 'list[L4Matcher]'


RuleMatch = TypedDict("PropertyMap", {"config": MatchConfig, "versionedExpr": str})


class OrganizationalPolicyRule(TypedDict):
    action: str
    match: RuleMatch
    priority: int
    description: str
    direction: str
    enableLogging: bool
    targetResources: 'list[str]'


class OrganizationPolicy(TypedDict):
    displayName: str
    type: str
    rules: 'list[OrganizationalPolicyRule]'


class Term(gcp.Term):
    """Used to create an individual term."""

    ACTION_MAP = {'accept': 'allow', 'next': 'goto_next'}

    _MAX_TERM_COMMENT_LENGTH = 64

    _TARGET_RESOURCE_FORMAT = (
        'https://www.googleapis.com/compute/v1/projects/{}/global/networks/{}'
    )

    _TERM_ADDRESS_LIMIT = 256

    _TERM_TARGET_RESOURCES_LIMIT = 256

    _TERM_DESTINATION_PORTS_LIMIT = 256

    def __init__(
        self,
        term: policy.Term,
        address_family: str = 'inet',
        policy_inet_version: str = 'inet',
        api_version: str = 'beta',
    ) -> None:
        super().__init__(term)
        self.address_family = address_family
        self.term = term
        self.skip = False
        self._ValidateTerm()
        self.api_version = api_version

        # This is to handle mixed, where the policy_inet_version is mixed,
        # but the term inet version is either inet/inet6.
        # This is only useful for term name and priority.
        self.policy_inet_version = policy_inet_version

    def _ValidateTerm(self) -> None:
        if self.term.destination_tag or self.term.source_tag:
            raise gcp.TermError('Hierarchical Firewall does not support tags')

        if len(self.term.target_resources) > self._TERM_TARGET_RESOURCES_LIMIT:
            raise gcp.TermError(
                'Term: %s  target_resources field contains %s resources. It should not contain more than "%s".'
                % (
                    self.term.name,
                    str(len(self.term.target_resources)),
                    self._TERM_TARGET_RESOURCES_LIMIT,
                )
            )

        for proj, vpc in self.term.target_resources:
            if not gcp.IsProjectIDValid(proj):
                raise gcp.TermError(
                    'Project ID "%s" must be 6 to 30 lowercase letters, digits, or hyphens.'
                    ' It must start with a letter. Trailing hyphens are prohibited.' % proj
                )
            if not gcp.IsVPCNameValid(vpc):
                raise gcp.TermError(
                    'VPC name "%s" must start with a lowercase letter '
                    'followed by up to 62 lowercase letters, numbers, '
                    'or hyphens, and cannot end with a hyphen.' % vpc
                )
        if self.term.source_port:
            raise gcp.TermError(
                'Hierarchical firewall does not support source port ' 'restrictions.'
            )
        if self.term.option:
            raise gcp.TermError(
                'Hierarchical firewall does not support the ' 'TCP_ESTABLISHED option.'
            )

        if len(self.term.destination_port) > self._TERM_DESTINATION_PORTS_LIMIT:
            raise gcp.TermError(
                'Term: %s destination_port field contains %s ports. It should not contain more than "%s".'
                % (
                    self.term.name,
                    str(len(self.term.destination_port)),
                    self._TERM_DESTINATION_PORTS_LIMIT,
                )
            )

        # Since policy_inet_version is used to handle 'mixed'.
        # We should error out if the individual term's inet version (address_family)
        # is anything other than inet/inet6, since this should never happen
        # naturally. Something has gone horribly wrong if you encounter this error.
        if self.address_family == 'mixed':
            raise gcp.TermError(
                'Hierarchical firewall rule has incorrect inet_version for rule: %s'
                % self.term.name
            )

    def ConvertToDict(self, priority_index: int) -> list[OrganizationPolicy]:
        """Converts term to dict representation of SecurityPolicy.Rule JSON format.

        Takes all of the attributes associated with a term (match, action, etc) and
        converts them into a dictionary which most closely represents
        the SecurityPolicy.Rule JSON format.

        Args:
          priority_index: An integer priority value assigned to the term.

        Returns:
          A dict term.
        """
        if self.skip:
            return {}

        rules = []

        # Identify if this is inet6 processing for a term under a mixed policy.
        mixed_policy_inet6_term = False
        if self.policy_inet_version == 'mixed' and self.address_family == 'inet6':
            mixed_policy_inet6_term = True

        term_dict = {
            'action': self.ACTION_MAP.get(self.term.action[0], self.term.action[0]),
            'direction': self.term.direction,
            'priority': priority_index,
        }

        # Get the correct syntax for API versions.
        src_ip_range = ApiVersionSyntaxMap.SYNTAX_MAP[self.api_version]['src_ip_range']
        dest_ip_range = ApiVersionSyntaxMap.SYNTAX_MAP[self.api_version]['dest_ip_range']
        layer_4_config = ApiVersionSyntaxMap.SYNTAX_MAP[self.api_version]['layer_4_config']

        target_resources = []
        for proj, vpc in self.term.target_resources:
            target_resources.append(self._TARGET_RESOURCE_FORMAT.format(proj, vpc))

        if target_resources:  # Only set when non-empty.
            term_dict['targetResources'] = target_resources

        term_dict['enableLogging'] = self._GetLoggingSetting()

        # This combo provides ability to identify the rule.
        term_name = self.term.name
        if mixed_policy_inet6_term:
            term_name = gcp.GetIpv6TermName(term_name)
        raw_description = f"{term_name}: {' '.join(self.term.comment)}"
        term_dict['description'] = gcp.TruncateString(
            raw_description, self._MAX_TERM_COMMENT_LENGTH
        )

        filtered_protocols = []
        for proto in self.term.protocol:
            # ICMP filtering by inet_version
            # Since each term has inet_version, 'mixed' is correctly processed here.
            if proto == 'icmp' and self.address_family == 'inet6':
                logging.warning(
                    'WARNING: Term %s is being rendered for inet6, ICMP '
                    'protocol will not be rendered.',
                    self.term.name,
                )
                continue
            if proto == 'icmpv6' and self.address_family == 'inet':
                logging.warning(
                    'WARNING: Term %s is being rendered for inet, ICMPv6 '
                    'protocol will not be rendered.',
                    self.term.name,
                )
                continue
            if proto == 'igmp' and self.address_family == 'inet6':
                logging.warning(
                    'WARNING: Term %s is being rendered for inet6, IGMP '
                    'protocol will not be rendered.',
                    self.term.name,
                )
                continue
            filtered_protocols.append(proto)
        # If there is no protocol left after ICMP/IGMP filtering, drop this term.
        # But only do this for terms that originally had protocols.
        # Otherwise you end up dropping the default-deny.
        if self.term.protocol and not filtered_protocols:
            return {}

        protocols_and_ports = []
        if not self.term.protocol:
            # Empty protocol list means any protocol, but any protocol in HF is
            # represented as "all"
            protocols_and_ports = [{'ipProtocol': 'all'}]
        else:
            for proto in filtered_protocols:
                # If the protocol name is not supported, use the protocol number.
                if proto not in self._ALLOW_PROTO_NAME:
                    proto = str(self.PROTO_MAP[proto])
                    logging.info(
                        'INFO: Term %s is being rendered using protocol number', self.term.name
                    )
                proto_ports = {'ipProtocol': proto}
                if self.term.destination_port:
                    ports = self._GetPorts()
                    if ports:  # Only set when non-empty.
                        proto_ports['ports'] = ports
                protocols_and_ports.append(proto_ports)

        if self.api_version == 'ga':
            term_dict['match'] = {layer_4_config: protocols_and_ports}
        else:
            term_dict['match'] = {'config': {layer_4_config: protocols_and_ports}}

        # match needs a field called versionedExpr with value FIREWALL
        # See documentation:
        # https://cloud.google.com/compute/docs/reference/rest/beta/organizationSecurityPolicies/addRule
        term_dict['match']['versionedExpr'] = 'FIREWALL'

        ip_version = self.AF_MAP[self.address_family]
        if ip_version == 4:
            any_ip = [nacaddr.IP('0.0.0.0/0')]
        else:
            any_ip = [nacaddr.IPv6('::/0')]

        if self.term.direction == 'EGRESS':
            daddrs = self.term.GetAddressOfVersion('destination_address', ip_version)

            # If the address got filtered out and is empty due to address family, we
            # don't render the term. At this point of term processing, the direction
            # has already been validated, so we can just log and return empty rule.
            if self.term.destination_address and not daddrs:
                logging.warning(
                    'WARNING: Term %s is not being rendered for %s, '
                    'because there are no addresses of that family.',
                    self.term.name,
                    self.address_family,
                )
                return []
            # This should only happen if there were no addresses set originally.
            if not daddrs:
                daddrs = any_ip

            destination_address_chunks = [
                daddrs[x : x + self._TERM_ADDRESS_LIMIT]
                for x in range(0, len(daddrs), self._TERM_ADDRESS_LIMIT)
            ]

            for daddr_chunk in destination_address_chunks:
                rule = copy.deepcopy(term_dict)
                if self.api_version == 'ga':
                    rule['match'][dest_ip_range] = [daddr.with_prefixlen for daddr in daddr_chunk]
                else:
                    rule['match']['config'][dest_ip_range] = [
                        daddr.with_prefixlen for daddr in daddr_chunk
                    ]
                rule['priority'] = priority_index
                rules.append(rule)
                priority_index += 1
        else:
            saddrs = self.term.GetAddressOfVersion('source_address', ip_version)

            # If the address got filtered out and is empty due to address family, we
            # don't render the term. At this point of term processing, the direction
            # has already been validated, so we can just log and return empty rule.
            if self.term.source_address and not saddrs:
                logging.warning(
                    'WARNING: Term %s is not being rendered for %s, '
                    'because there are no addresses of that family.',
                    self.term.name,
                    self.address_family,
                )
                return []
            # This should only happen if there were no addresses set originally.
            if not saddrs:
                saddrs = any_ip

            source_address_chunks = [
                saddrs[x : x + self._TERM_ADDRESS_LIMIT]
                for x in range(0, len(saddrs), self._TERM_ADDRESS_LIMIT)
            ]
            for saddr_chunk in source_address_chunks:
                rule = copy.deepcopy(term_dict)
                if self.api_version == 'ga':
                    rule['match'][src_ip_range] = [saddr.with_prefixlen for saddr in saddr_chunk]
                else:
                    rule['match']['config'][src_ip_range] = [
                        saddr.with_prefixlen for saddr in saddr_chunk
                    ]
                rule['priority'] = priority_index
                rules.append(rule)
                priority_index += 1

        return rules

    def __str__(self) -> str:
        return ''


class HierarchicalFirewall(gcp.GCP):
    """A GCP Hierarchical Firewall policy."""

    SUFFIX = '.gcphf'
    _ANY_IP = {
        'inet': nacaddr.IP('0.0.0.0/0'),
        'inet6': nacaddr.IP('::/0'),
    }
    _PLATFORM = 'gcp_hf'
    _SUPPORTED_AF = frozenset(['inet', 'inet6', 'mixed'])
    # Beta is the default API version. GA supports IPv6 (inet6/mixed).
    _SUPPORTED_API_VERSION = frozenset(['beta', 'ga'])
    _DEFAULT_MAXIMUM_COST = 100

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          Tuple containing both supported tokens and sub tokens.
        """
        supported_tokens, _ = super()._BuildTokens()

        supported_tokens |= {
            'destination_tag',
            'expiration',
            'source_tag',
            'translated',
            'target_resources',
            'logging',
        }

        supported_tokens -= {
            'destination_address_exclude',
            'expiration',
            'icmp_type',
            'source_address_exclude',
            'verbatim',
        }

        supported_sub_tokens = {'action': {'accept', 'deny', 'next'}}
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        """Translates a Aerleon policy into a HF-specific data structure.

        Takes in a POL file, parses each term and populates the policy
        dict. Each term in this list is a dictionary formatted according to
        HF's rule API specification.  Additionally, checks for its quota.

        Args:
          pol: A Policy() object representing a given POL file.
          exp_info: An int that specifies number of weeks until policy expiry.

        Raises:
          ExceededCostError: Raised when the cost of a policy exceeds the default
              maximum cost.
          HeaderError: Raised when the header cannot be parsed or a header option is
              invalid.
          DifferentPolicyNameError: Raised when a header policy name differs from
              other in the same policy.
        """
        self.policies = []
        policy = {'rules': [], 'type': 'FIREWALL'}
        is_policy_modified = False
        counter = 1
        total_cost = 0
        policies_max_cost = self._DEFAULT_MAXIMUM_COST
        previous_max_cost = -1
        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)

            is_policy_modified = True

            # Get term direction if set.
            direction = 'INGRESS'
            for i in self._GOOD_DIRECTION:
                if i in filter_options:
                    direction = i
                    filter_options.remove(i)

            # Get the address family if set.
            address_family = 'inet'
            for i in self._SUPPORTED_AF:
                if i in filter_options:
                    address_family = i
                    filter_options.remove(i)

            # Get the compute API version if set.
            api_version = 'beta'
            for i in self._SUPPORTED_API_VERSION:
                if i in filter_options:
                    api_version = i
                    filter_options.remove(i)
                    break

            # Find the default maximum cost of a policy, an integer, if specified.
            max_cost = self._DEFAULT_MAXIMUM_COST
            for opt in filter_options:
                try:
                    max_cost = int(opt)
                    filter_options.remove(opt)
                    break
                except ValueError:
                    continue

            if max_cost > 65536:
                raise gcp.HeaderError('Default maximum cost cannot be higher than 65536')

            if previous_max_cost != -1 and previous_max_cost != max_cost:
                raise gcp.HeaderError(
                    'Maximum costs of each policy specified must be equal. '
                    'Unequal costs found: %d and %d' % (previous_max_cost, max_cost)
                )

            policies_max_cost = max_cost
            previous_max_cost = max_cost

            display_name = ApiVersionSyntaxMap.SYNTAX_MAP[api_version]['display_name']

            # Get policy name and validate it to meet displayName requirements.
            policy_name = header.FilterName(self._PLATFORM)
            if not policy_name:
                raise gcp.HeaderError('Policy name was not specified in header')
            filter_options.remove(policy_name)
            if len(policy_name) > 63:
                raise gcp.HeaderError(
                    'Policy name "%s" is too long; the maximum number of characters '
                    'allowed is 63' % (policy_name)
                )
            if not bool(re.match('^[a-z]([-a-z0-9]*[a-z0-9])?$', policy_name)):
                raise gcp.HeaderError(
                    'Invalid string for displayName, "%s"; the first character must be '
                    'a lowercase letter, and all following characters must be a dash, '
                    'lowercase letter, or digit, except the last character, which '
                    'cannot be a dash.' % (policy_name)
                )
            if display_name in policy and policy[display_name] != policy_name:
                raise DifferentPolicyNameError(
                    'Policy names that are from the same policy are expected to be '
                    'equal, but %s is different to %s' % (policy[display_name], policy_name)
                )
            policy[display_name] = policy_name

            # If there are remaining options, they are unknown/unsupported options.
            if filter_options:
                raise gcp.HeaderError(
                    'Unsupported or unknown filter options %s in policy %s '
                    % (str(filter_options), policy_name)
                )

            # Handle mixed for each indvidual term as inet and inet6.
            # inet/inet6 are treated the same.
            term_address_families = []
            if address_family == 'mixed':
                term_address_families = ['inet', 'inet6']
            else:
                term_address_families = [address_family]

            for term in terms:
                if term.stateless_reply:
                    continue

                if gcp.IsDefaultDeny(term):
                    if direction == 'EGRESS':
                        if address_family != 'mixed':
                            # Default deny also gets processed as part of terms processing.
                            # The name and priority get updated there.
                            term.destination_address = [self._ANY_IP[address_family]]
                        else:
                            term.destination_address = [
                                self._ANY_IP['inet'],
                                self._ANY_IP['inet6'],
                            ]
                    else:
                        if address_family != 'mixed':
                            term.source_address = [self._ANY_IP[address_family]]
                        else:
                            term.source_address = [self._ANY_IP['inet'], self._ANY_IP['inet6']]
                term.name = self.FixTermLength(term.name)
                term.direction = direction

                for term_af in term_address_families:
                    rules = Term(
                        term,
                        address_family=term_af,
                        policy_inet_version=address_family,
                        api_version=api_version,
                    ).ConvertToDict(priority_index=counter)
                    if not rules:
                        continue
                    for dict_term in rules:
                        total_cost += GetRuleTupleCount(dict_term, api_version)
                        policy['rules'].append(dict_term)
                    counter += len(rules)

        # We want to check the total policy cost, not just per policy.
        if total_cost > policies_max_cost:
            raise ExceededCostError(
                'Policy cost (%d) for %s reached the '
                'maximum (%d)' % (total_cost, policy[display_name], policies_max_cost)
            )

        self.policies.append(policy)

        # Do not render an empty rules if no policies have been evaluated.
        if not is_policy_modified:
            self.policies = []

        if total_cost > 0:
            logging.info('Policy %s quota cost: %d', policy[display_name], total_cost)


def GetRuleTupleCount(dict_term: dict[str, Union[list, str]], api_version: str) -> int:
    """Calculate the tuple count of a rule in its dictionary form.

    Quota is charged based on how complex the rules are rather than simply
    limiting the number of rules.

    The cost of a rule is the number of distinct protocol:port combinations plus
    the number of IP addresses plus the number of targets.

    Note: The goal of this function is not to determine if a rule is valid, but
        to calculate its tuple count regardless of correctness.

    Args:
      dict_term: A dict object.
      api_version: A string indicating the api version.

    Returns:
      int: The tuple count of the rule.
    """
    layer4_count = 0
    layer_4_config = ApiVersionSyntaxMap.SYNTAX_MAP[api_version]['layer_4_config']
    dest_ip_range = ApiVersionSyntaxMap.SYNTAX_MAP[api_version]['dest_ip_range']
    src_ip_range = ApiVersionSyntaxMap.SYNTAX_MAP[api_version]['src_ip_range']
    targets_count = len(dict_term.get('targetResources', []))
    if api_version == 'ga':
        config = dict_term.get('match', {})
    else:
        config = dict_term.get('match', {}).get('config', {})

    addresses_count = len(config.get(dest_ip_range, []) + config.get(src_ip_range, []))

    for l4config in config.get(layer_4_config, []):
        for _ in l4config.get('ports', []):
            layer4_count += 1
        if l4config.get('ipProtocol'):
            layer4_count += +1

    return addresses_count + layer4_count + targets_count



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Speedway iptables generator.  This is a subclass of Iptables lib."""

from string import Template

from aerleon.lib import iptables


class Error(iptables.Error):
    pass


class Term(iptables.Term):
    """Generate Iptables policy terms."""

    _PLATFORM = 'speedway'
    _PREJUMP_FORMAT = None
    _POSTJUMP_FORMAT = Template('-A $filter -j $term')


class Speedway(iptables.Iptables):
    """Generates filters and terms from provided policy object."""

    _PLATFORM = 'speedway'
    _DEFAULT_PROTOCOL = 'all'
    SUFFIX = '.ipt'

    _RENDER_PREFIX = '*filter'
    _RENDER_SUFFIX = 'COMMIT'
    _DEFAULTACTION_FORMAT = ':%s %s [0:0]'

    _TERM = Term



# Copyright 2022 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""NFtables policy generator for Aerleon."""

import collections
import copy
from typing import DefaultDict, Union

from absl import logging

from aerleon.lib import aclgenerator, nacaddr, policy

# NFTables and Aerleon have conflicting definitions of 'address family'
# In Aerleon:
# 'mixed' refers to a 'mixed address policy IPv4/IPv6'
# 'inet6' refers to IPv6 only.
# 'inet' refers to IPv4 only.
# In nftables:
# 'inet' refers to mixed IPv4/IPv6 policies.
# 'ip6' IPv6 only.
# 'ip' IPv4 only.
# Therefore; we use static global variables in this generator to refer to the
# real intent, values are the NFtable AF format.
ip4 = 'ip'
ip6 = 'ip6'
mixed = 'inet'


def TabSpacer(number_spaces: int, string: str) -> str:
    """Configuration indentation utility function."""
    blank_space = ' '
    return (blank_space * number_spaces) + string


def Add(statement: str) -> str:
    """Prefix space appending utility to handle text joins."""
    if statement:
        return TabSpacer(1, statement)
    else:
        return statement


def ChainFormat(kind: str, name: str, ruleset: list[str]) -> str:
    """Builds a chain in NFtables configuration format.

    Args:
      kind: type string (chain or counter)
      name: name to give the chain.
      ruleset: the list returned from RulesetGenerator function.

    Returns:
      chain_strings: multi-line string nftable configuration for the chain.
    """
    header_sp = 4
    content_sp = 8
    chain_output = []
    chain_output.append(TabSpacer(header_sp, f'{kind} {name} {{'))
    for line in ruleset:
        chain_output.append(TabSpacer(content_sp, line))
    chain_output.append(TabSpacer(header_sp, '}'))
    return '\n'.join(chain_output)


class Error(aclgenerator.Error):
    """Base error class."""


class TermError(Error):
    """Raised when a term is not valid."""


class HeaderError(Error):
    """Raised when a header is not valid."""


class UnsupportedFilterTypeError(Error):
    """Raised when an unsupported filter type is specified."""


class UnsupportedExpressionError(Error):
    """Raised when an unsupported expression is specified."""


class Term(aclgenerator.Term):
    """Representation of an individual NFT term.

    This is primarily useful for Nftables.__str__() method.

    Args: term policy.Term object
    """

    _ALLOWED_PROTO_NAME = frozenset(
        ['tcp', 'udp', 'icmp', 'esp', 'udp', 'ah', 'comp', 'udplite', 'dccp', 'sctp', 'icmpv6']
    )
    _ACTIONS = {'accept': 'accept', 'deny': 'drop'}

    def __init__(self, term: policy.Term, nf_af: str, nf_hook: str, verbose: bool = True):
        """Individual instances of a Term for NFtables.

        Args:
          term: Term data.
          nf_af: nftables table type IPv4 only (ip), IPv6 (ip6) or dual-stack
            (inet).
          nf_hook: INPUT or OUTPUT (packet processing/direction of traffic).
          verbose: used for comment handling.
        """
        self.term = term
        self.address_family = nf_af
        self.hook = nf_hook
        self.verbose = verbose

    def MapICMPtypes(self, af: SyntaxWarning, term_icmp_types: list[str]) -> list[str]:
        """Normalize certain ICMP_TYPES for NFTables rendering.

        If we encounter certain keyword values in policy.Term.ICMP_TYPE keywords,
        we override them with NFTable specific values in order for rendered
        policy to be semantically correct with what NFT expects.
        https://www.netfilter.org/projects/nftables/manpage.html

        Function is used inside PortsAndProtocols.

        Args:
          af:
          term_icmp_types: ICMP types keywords.

        Returns:
          normalized list of icmp_types.
        """
        ICMP_TYPE_REMAP = {
            6: {
                'multicast-listener-query': 'mld-listener-query',
                'multicast-listener-report': 'mld-listener-report',
                'multicast-listener-done': 'mld-listener-done',
                'router-solicit': 'nd-router-solicit',
                'router-advertisement': 'nd-router-advert',
                'neighbor-solicit': 'nd-neighbor-solicit',
                'neighbor-advertisement': 'nd-neighbor-advert',
                'redirect-message': 'nd-redirect',
                'inverse-neighbor-discovery-solicitation': 'ind-neighbor-solicit',
                'inverse-neighbor-discovery-advertisement': 'ind-neighbor-advert',
                'version-2-multicast-listener-report': 'mld2-listener-report',
            },
            4: {
                # IPv4 exceptions below
                'unreachable': 'destination-unreachable',
                'information-request': 'info-request',
                'information-reply': 'info-reply',
                'mask-request': 'address-mask-request',
                'mask-reply': 'address-mask-reply',
            },
        }

        for item in term_icmp_types:
            if af == ip4:
                # IPv4 ICMP
                if item in ICMP_TYPE_REMAP[4]:
                    # Replace with NFT expected value.
                    term_icmp_types[term_icmp_types.index(item)] = ICMP_TYPE_REMAP[4].get(item)
            if af == ip6:
                # IPv6 ICMP
                if item in ICMP_TYPE_REMAP[6]:
                    # Replace with NFT expected value.
                    term_icmp_types[term_icmp_types.index(item)] = ICMP_TYPE_REMAP[6].get(item)
        return term_icmp_types

    def CreateAnonymousSet(self, data: list[str]) -> str:
        """Build a nftables anonymous set from some elements.

        Anonymous are formatted using curly braces then some data. These sets are
        bound to a rule, have no specific name and cannot be updated.

        Args:
          data: a list of strings to format.

        Returns:
          formatted string of items as anonymous set.
        """
        nfset = []
        if isinstance(data, str):
            # Handle single string. No params.
            return data
        if len(data) == 1:
            # Handle a list of a single element.
            nfset = data[0]
            return nfset
        if len(data) > 1:
            nfset = ', '.join(data)
            return f'{{ {nfset} }}'

    def PortsAndProtocols(self, address_family, protocol, src_ports, dst_ports, icmp_type):
        """Handling protocol specific NFTable statements.

        Args:
          address_family: term address family.
          protocol: term protocol.
          src_ports: raw term source port.
          dst_ports: raw term dest port.
          icmp_type: special ICMP type flag.

        Returns:
          list of statements related to ports and protocols.

        """

        def PortStatement(protocol: list[str], source: str, destination: str) -> list[str]:
            """NFT port statement. Returns empty if no ports defined."""
            ports_list = []

            # SOURCE PORTS.
            if source:
                ports_list.append(f'{protocol} sport {self.CreateAnonymousSet(source)}')

            # DESTINATION PORTS.
            if destination:
                ports_list.append(f'{protocol} dport {self.CreateAnonymousSet(destination)}')

            # Normalize ports into single nft statement.
            if ports_list:
                ports_statement = ' '.join(ports_list)
            else:
                ports_statement = ''
            return ports_statement  # end PortStatement.

        ip_protocol = copy.deepcopy(protocol)
        ip6_protocol = copy.deepcopy(protocol)
        # Normalize term.ports objects.
        src_p = self._Group(src_ports)
        dst_p = self._Group(dst_ports)
        statement_lines = []

        # Normalize ICMP types.
        # TODO: Call self.NormalizeIcmpTypes.
        icmp_type = self.MapICMPtypes(address_family, icmp_type)

        if address_family == mixed:
            # The way we handle mixed is we call ourselves twice.
            ipv4_list = self.PortsAndProtocols(ip4, protocol, src_ports, dst_ports, icmp_type)
            ipv6_list = self.PortsAndProtocols(ip6, protocol, src_ports, dst_ports, icmp_type)
            return ipv4_list + ipv6_list

        if address_family == 'ip':
            # IPv4 stuff.
            if icmp_type and ('icmp' in ip_protocol):
                if len(icmp_type) > 1:
                    statement_lines.append(f"icmp type{Add(self.CreateAnonymousSet(icmp_type))}")
                else:
                    statement_lines.append(f"icmp type{Add(icmp_type)}")
                ip_protocol.remove('icmp')
            if 'icmpv6' in ip_protocol:
                # No IPv6 protocols in IPv4 family.
                ip_protocol.remove('icmpv6')
            if ip_protocol:
                # Multi-protocol and zero-ports.
                if len(ip_protocol) > 1 and not (src_ports or dst_ports):
                    statement_lines.append(
                        f"ip protocol{Add(self.CreateAnonymousSet(ip_protocol))}"
                    )
                else:
                    for proto in ip_protocol:
                        if src_ports or dst_ports:
                            statement_lines.append(PortStatement(proto, src_p, dst_p))
                        else:
                            statement_lines.append(f"ip protocol{Add(proto)}")

        if address_family == 'ip6':
            # IPv6 stuff.
            if icmp_type and ('icmpv6' in ip6_protocol):
                if len(icmp_type) > 1:
                    statement_lines.append(f"icmpv6 type{Add(self.CreateAnonymousSet(icmp_type))}")
                else:
                    statement_lines.append(f"icmpv6 type{Add(icmp_type)}")
                ip6_protocol.remove('icmpv6')
            if 'icmp' in ip6_protocol:
                # No IPv4 protocols in IPv6 family.
                ip6_protocol.remove('icmp')
            if ip6_protocol:
                # NFT IPv6 protocol matching is complex. Using 'ip6 nexthdr' only
                # matches if ipv6 packet does not contain any extension headers.
                # we use meta l4proto here to walk down the headers until real transport
                # protocol is found. This allows us to use Sets here too.
                # https://wiki.nftables.org/wiki-nftables/index.php/Matching_packet_headers
                if len(ip6_protocol) > 1 and not (src_ports or dst_ports):
                    statement_lines.append(
                        f"meta l4proto{Add(self.CreateAnonymousSet(ip6_protocol))}"
                    )
                else:
                    # We avoid using th (transport header), instead we use single
                    # statements for each protocol.
                    for proto in ip6_protocol:
                        if src_ports or dst_ports:
                            statement_lines.append(PortStatement(proto, src_p, dst_p))
                        else:
                            # Single proto, no ports.
                            statement_lines.append(f"meta l4proto{Add(proto)}")

        return statement_lines

    def _OptionsHandler(self, term: policy.Term) -> str:
        """Term 'option' handler.

        Function used to evaluate term.logging and also term.option values. Then
        it builds any statement that would be appended before a veredict.
        Results of this function are then used in GroupExpressions() to combine
        a final valid NFTables chain.

        Args:
          term: Aerleon Term data.

        Returns:
          list of statements related to generator options.
        """
        options = []

        # Stateful firewall, Accept only NEW traffic for the specific term.
        # Base chain already allows all return traffic of
        # state (ESTABLISHED, RELATED)
        # This should prevent invalid, untracked packets from being accepted.
        if 'deny' not in term.action:
            options.append('ct state new')

        # 'logging' handling.
        if term.logging:
            # str() trick to circumvent VarType class attr comparison checks.
            if 'disable' not in str(term.logging):
                # Simple syslogging implementation.
                options.append(f'log prefix "{term.name}"')

        # 'counter' handling.
        # https://wiki.nftables.org/wiki-nftables/index.php/Counters
        # We don't use named counters here because we already structure NFT ruleset
        # in child chains per each rule. So simply looking at term_child_chain is
        # easy to tell the counter stats for that ruleset.
        if term.counter:
            options.append('counter')

        # Build the final statement to be returned.
        if options:
            return ' '.join(options)
        else:
            return ''

    def GroupExpressions(
        self, address_expr: list[str], pp_expr: list[str], options: str, verdict: str
    ) -> list[str]:
        """Combines all expressions with a verdict (decision).

        The inputs are already pre-sanitized by RulesetGenerator. NFTables processes
        rules from left-to-right - ending in a verdict. We form our ruleset then
        towards the end append any term.options from _OptionsHandler.

        Args:
          address_expr: pre-processed list of nftable statements of network
            addresses.
          pp_expr: pre-processed list of nftables protocols and ports.
          options: string value to append before verdict for NFT special options.
          verdict: action to take on resulting final statement (allow/deny).

        Returns:
          list of strings representing valid nftables statements.
        """
        statement = []
        if address_expr:
            for addr in address_expr:
                if pp_expr:
                    for pstat in pp_expr:
                        if pstat.startswith('icmp type') or addr.startswith('ip '):
                            # Handle IPv4 ports and proto statements.
                            if addr.startswith('ip '):
                                statement.append(addr + Add(pstat) + Add(options) + Add(verdict))
                        elif pstat.startswith('icmpv6 type') or addr.startswith('ip6'):
                            if addr.startswith('ip6'):
                                statement.append(addr + Add(pstat) + Add(options) + Add(verdict))
                else:
                    statement.append(addr + Add(options) + Add(verdict))
        elif pp_expr:
            # Handle statement without addresses but has ports & protocols.
            for pstat in pp_expr:
                statement.append(pstat + Add(options) + Add(verdict))
        else:
            # If no addresses or ports & protocol. Verdict only statement.
            statement.append(Add(options) + verdict)
        return statement

    def _AddrStatement(
        self,
        address_family: str,
        src_addr: list[Union[nacaddr.IPv4, nacaddr.IPv6]],
        dst_addr: list[Union[nacaddr.IPv4, nacaddr.IPv6]],
    ) -> list[str]:
        """Builds an NFTables address statement.

        Args:
          address_family: NFTables address family.
          src_addr: prefiltered list of src addresses.
          dst_addr: prefiltered list of dst addresses.

        Returns:
          list of strings representing valid nftables address statements (IPv4/6).
        """
        address_statement = []
        src_addr_book = self._AddressClassifier(src_addr)
        dst_addr_book = self._AddressClassifier(dst_addr)

        if src_addr and dst_addr:
            # Condition where term has both defined.
            if address_family == 'inet' or address_family == 'ip':
                if src_addr_book['ip'] and dst_addr_book['ip']:
                    address_statement.append(
                        'ip saddr '
                        + self.CreateAnonymousSet(src_addr_book['ip'])
                        + ' '
                        + 'ip daddr '
                        + self.CreateAnonymousSet(dst_addr_book['ip'])
                    )
            if address_family == 'inet' or address_family == 'ip6':
                if src_addr_book['ip6'] and dst_addr_book['ip6']:
                    address_statement.append(
                        'ip6 saddr '
                        + self.CreateAnonymousSet(src_addr_book['ip6'])
                        + ' '
                        + 'ip6 daddr '
                        + self.CreateAnonymousSet(dst_addr_book['ip6'])
                    )
        elif src_addr:
            # Term has only src defined.
            if address_family == 'inet' or address_family == 'ip':
                if src_addr_book['ip']:
                    address_statement.append(
                        f"ip saddr {self.CreateAnonymousSet(src_addr_book['ip'])}"
                    )
            if address_family == 'inet' or address_family == 'ip6':
                if src_addr_book['ip6']:
                    address_statement.append(
                        f"ip6 saddr {self.CreateAnonymousSet(src_addr_book['ip6'])}"
                    )
        elif dst_addr:
            if address_family == 'inet' or address_family == 'ip':
                if dst_addr_book['ip']:
                    address_statement.append(
                        f"ip daddr {self.CreateAnonymousSet(dst_addr_book['ip'])}"
                    )
            if address_family == 'inet' or address_family == 'ip6':
                if dst_addr_book['ip6']:
                    address_statement.append(
                        f"ip6 daddr {self.CreateAnonymousSet(dst_addr_book['ip6'])}"
                    )
        return address_statement

    def RulesetGenerator(self, term: policy.Term) -> list[str]:
        """Generate string rules of a given Term.

        Rules are constructed from Terms() and are contained within chains.
        This function generates rules that will be present inside a regular
        (non-base) chain. Each item in list represents a line break for later
        parsing.

        Args:
          term: term data.

        Returns:
          list of strings. Representing a ruleset for later formatting.

        """
        term_ruleset = []

        address_list = []
        # COMMENT handling.
        if self.verbose:
            for line in self.term.comment:
                term_ruleset.append(f'comment "{line}"')

        # ADDRESS handling.
        address_list = self._AddrStatement(
            self.address_family, self.term.source_address, self.term.destination_address
        )

        # PORTS and PROTOCOLS handling.
        proto_and_ports = self.PortsAndProtocols(
            self.address_family,
            self.term.protocol,
            self.term.source_port,
            self.term.destination_port,
            self.term.icmp_type,
        )

        # OPTIONS / LOGGING / COUNTERS
        opt = self._OptionsHandler(term)
        # STATEMENT VERDICT / ACTION.
        verdict = self._ACTIONS[self.term.action[0]]
        # TODO: If verdict is not supported, drop nftable_rule for it.
        nftable_rule = self.GroupExpressions(address_list, proto_and_ports, opt, verdict)
        term_ruleset.extend(nftable_rule)
        return term_ruleset

    def _AddressClassifier(
        self, address_to_classify: list[Union[nacaddr.IPv4, nacaddr.IPv6]]
    ) -> DefaultDict[str, list[Union[nacaddr.IPv4, nacaddr.IPv6]]]:
        """Organizes network addresses according to IP family in a dict.

        Args:
          address_to_classify: nacaddr.IP list of network addresses.

        Returns:
          dictionary of network addresses classified by AF.
        """
        addresses = collections.defaultdict(list)
        for addr in address_to_classify:
            if addr.version == 4:
                addresses['ip'].append(str(addr))
            if addr.version == 6:
                addresses['ip6'].append(str(addr))
        return addresses

    def _Group(self, group: list[tuple[int, int]]) -> str:
        """If 1 item return it, else return [ item1 item2 ].

        Args:
          group: a list.  could be a list of strings (protocols) or a list of tuples
            (ports)

        Returns:
          rval: a string surrounded by '[' and ']'
        """

        def _FormatPorts(port):
            if isinstance(port, int):
                return str(port)
            elif port[0] == port[1]:
                return '%d' % port[0]
            else:
                # port range
                return '%d-%d' % (port[0], port[1])

        if len(group) > 1:
            rval = [_FormatPorts(x) for x in group]
        elif len(group) == 1:
            rval = _FormatPorts(group[0])
        else:
            # Ports undefined/empty.
            rval = ''
        return rval

    def __str__(self) -> str:
        """Terms printing function.

        Each term is expressed as its own chain. Later referenced to a parent chain
        with filter directionality (input/output).
        """
        return ChainFormat('chain', self.term.name, self.RulesetGenerator(self.term))


class Nftables(aclgenerator.ACLGenerator):
    """A NFtables policy object."""

    _PLATFORM = 'nftables'
    SUFFIX = '.nft'
    _HEADER_AF = frozenset(('inet', 'inet6', 'mixed'))
    _SUPPORTED_HOOKS = frozenset(('input', 'output'))
    _HOOK_PRIORITY_DEFAULT = 0
    _BASE_CHAIN_PREFIX = 'root'
    _LOGGING = set()

    _OPTIONAL_SUPPORTED_KEYWORDS = frozenset(
        [
            'expiration',
        ]
    )

    _AF_MAP = {'inet': (4,), 'inet6': (6,), 'mixed': (4, 6)}
    # Below mapping converts Aerleon HEADER native to nftables table.
    # In Nftables 'inet' contains both IPv4 and IPv6 addresses and rules.
    NF_TABLE_AF_MAP = {'inet': 'ip', 'inet6': 'ip6', 'mixed': 'inet'}

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """NFTables generator list of supported tokens and sub tokens.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()
        # Set of supported keywords for a given platform.  Values should be in
        # undercase form, eg, icmp_type (not icmp-type)
        supported_tokens = {
            'action',
            'comment',
            'destination_address',
            'destination_address_exclude',
            'destination_port',
            'expiration',
            'icmp_type',
            'name',  # obj attribute, not token
            'option',
            'protocol',
            'platform',
            'platform_exclude',
            'source_address',
            'source_address_exclude',
            'source_port',
            'translated',  # obj attribute, not token
            'stateless_reply',
        }

        # These keys must be also listed in supported_tokens.
        # Keys should be in undercase form, eg, icmp_type (not icmp-type). Values
        # should be in dash form, icmp-type (not icmp_type)
        supported_sub_tokens = {
            'option': {
                'established',
                'tcp-established',
            },
            'action': {
                'accept',
                'deny',
            },
            'icmp_type': set(list(Term.ICMP_TYPE[4].keys()) + list(Term.ICMP_TYPE[6].keys())),
        }
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int):
        """Translates a Aerleon policy file into NFtables specific data structure.

        Reads a POL file, filters for NFTables specific data, parses each term
        and populates the nftables_policies list.

        Args:
          pol: A Policy() object representing a given POL file.
          exp_info: An int that specifies number of weeks until policy expires.

        Raises:
          TermError: Raised when policy term requirements are not met.
        """
        self.nftables_policies = []

        pol_counter = 0

        for header, terms in pol.filters:
            filter_options = header.FilterOptions('nftables')
            (
                nf_af,
                nf_hook,
                nf_priority,
                filter_policy_default_action,
                verbose,
            ) = self._ProcessHeader(filter_options)

            # Base chain determine name based on iteration of header.
            base_chain_name = self._BASE_CHAIN_PREFIX + str(pol_counter)
            child_chains = collections.defaultdict(dict)
            term_names = set()
            new_terms = []
            # TODO: Add checks for ICMP and address families.
            for term in terms:
                if term.name in term_names:
                    raise TermError('Duplicate term name')
                term_names.add(term.name)
                if term.stateless_reply:
                    logging.warning(
                        'WARNING: Term %s is a stateless reply ' 'term and will not be rendered.',
                        term.name,
                    )
                    continue
                # This generator is stateful, we don't do stateless rules.
                # Stateful firewalls don't require a reverse rule/term; thus skip.
                if 'established' in term.option:
                    logging.warning(
                        'WARNING: Term %s is a established ' 'term and will not be rendered.',
                        term.name,
                    )
                    continue
                if 'tcp-established' in term.option:
                    logging.warning(
                        'WARNING: Term %s is a tcp-established ' 'term and will not be rendered.',
                        term.name,
                    )
                    continue
                # Handle address excludes before building nft address book dict.
                for i in term.source_address_exclude:
                    term.source_address = nacaddr.RemoveAddressFromList(term.source_address, i)
                for i in term.destination_address_exclude:
                    term.destination_address = nacaddr.RemoveAddressFromList(
                        term.destination_address, i
                    )
                new_terms.append(Term(term, nf_af, nf_hook, verbose))
                # Instantiate object to call function from Term()
                term_object = Term(term, nf_af, nf_hook, verbose)
                child_chains[base_chain_name].update(
                    {term.name: term_object.RulesetGenerator(term)}
                )
            pol_counter += 1
            self.nftables_policies.append(
                (
                    header,
                    base_chain_name,
                    nf_af,
                    nf_hook,
                    nf_priority,
                    filter_policy_default_action,
                    verbose,
                    child_chains,
                )
            )

    def _ProcessHeader(self, header_options: list[str]) -> tuple[str, str, int, str, bool]:
        """Aerleon policy header processing.

        Args:
          header_options: Aerleon policy header data (filter_options)

        Raises:
          HeaderError: Raised when the policy header format requirements are not
          met.

        Returns:
          netfilter_family: x. filter_options[0]
          netfilter_hook: x. filter_options[1].lower()
          netfilter_priority: numbers = [x for x in filter_options if x.isdigit()]
          policy_default_action: nftable action to take on unmatched packets.
          verbose: header and term verbosity.
        """
        if len(header_options) < 2:
            raise HeaderError('Invalid header for Nftables. Required fields missing.')
        # First header element should dictate type of policy.
        if header_options[0] not in Nftables._HEADER_AF:
            raise HeaderError(
                'Invalid address family in header: %s. Supported: %s'
                % (header_options[0], Nftables._HEADER_AF)
            )
        netfilter_family = self.NF_TABLE_AF_MAP.get(header_options[0])
        policy_default_action = 'drop'
        if 'ACCEPT' in header_options:
            policy_default_action = 'accept'
        netfilter_hook = header_options[1].lower()
        if netfilter_hook not in self._SUPPORTED_HOOKS:
            raise HeaderError(
                '%s is not a supported nftables hook. Supported hooks: %s'
                % (netfilter_hook, list(self._SUPPORTED_HOOKS))
            )
        if len(header_options) >= 2:
            numbers = [x for x in header_options if x.isdigit()]
            if not numbers:
                netfilter_priority = self._HOOK_PRIORITY_DEFAULT
                logging.info(
                    'INFO: NFtables priority not specified in header.' 'Defaulting to %s',
                    self._HOOK_PRIORITY_DEFAULT,
                )
            if len(numbers) == 1:
                # A single integer value is used to set priority.
                netfilter_priority = numbers[0]
            if len(numbers) > 1:
                raise HeaderError('Too many integers in header.')
        verbose = True
        if 'noverbose' in header_options:
            verbose = False
            header_options.remove('noverbose')
        return netfilter_family, netfilter_hook, netfilter_priority, policy_default_action, verbose

    def _ConfigurationDictionary(
        self, nft_pol: list[Union[policy.Header, str, int, bool, DefaultDict]]
    ) -> DefaultDict[str, Union[str, DefaultDict]]:
        """NFTables configuration object.

        Organizes policies into a data structure that can keep relationships with
        NFTables address family (tables) and the parent base chain (+ child chains).

        Args:
          nft_pol: Object containing pre-processed data from _TranslatePolicy.

        Returns:
          nftables: dictionary of dictionaries NFTables policy object.
        """
        nftables = collections.defaultdict(dict)
        for (
            header,
            base_chain_name,
            nf_af,
            nf_hook,
            nf_priority,
            filter_policy_default_action,
            verbose,
            child_chains,
        ) in nft_pol:
            base_chain_comment = ''
            # Add max character checking on header.comment later if needed.
            if verbose:
                base_chain_comment = header.comment
            nftables[nf_af][base_chain_name] = {
                'hook': nf_hook,
                'comment': base_chain_comment,
                'priority': nf_priority,
                'policy': filter_policy_default_action,
                'rules': child_chains,
            }
        return nftables

    def __str__(self) -> str:
        """Render the policy as Nftables configuration."""
        nft_config = []
        configuration = self._ConfigurationDictionary(self.nftables_policies)

        for address_family in configuration:
            nft_config.append('table %s filtering_policies {' % address_family)
            base_chain_dict = configuration[address_family]
            for item in base_chain_dict:
                # TODO: If we ever add NFTables 'named counters' it would go here.
                for k, v in base_chain_dict[item]['rules'][item].items():
                    nft_config.append(ChainFormat('chain', k, v))
                # base chain header and contents.
                nft_config.append(TabSpacer(4, 'chain %s {' % item))
                if base_chain_dict[item]['comment']:
                    # Handle multi-line comments
                    for comment in base_chain_dict[item]['comment']:
                        nft_config.append(TabSpacer(8, f'comment "{comment}"'))
                nft_config.append(
                    TabSpacer(
                        8,
                        'type filter hook %s priority %s; policy %s;'
                        % (
                            base_chain_dict[item]['hook'],
                            base_chain_dict[item]['priority'],
                            base_chain_dict[item]['policy'],
                        ),
                    )
                )
                # stateful firewall: allow reply traffic.
                nft_config.append(TabSpacer(8, 'ct state established,related accept'))
                # Reference the child chains with jump.
                for child_chain in base_chain_dict[item]['rules'][item].keys():
                    nft_config.append(TabSpacer(8, f'jump {child_chain}'))
                nft_config.append(TabSpacer(4, '}'))  # chain_end
            nft_config.append('}')  # table_end

        # Terminating newline.
        nft_config.append('\n')
        return '\n'.join(nft_config)



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Check where hosts, ports and protocols are matched in an Aerleon policy."""

import logging
from collections import defaultdict

from aerleon.lib import nacaddr, naming, policy, policy_builder, port


class Error(Exception):
    """Base error class."""


class AddressError(Error):
    """Incorrect IP address or format."""


class BadPolicyError(Error):
    """Item is not a valid policy object."""


class NoTargetError(Error):
    """Specified target platform not available in specified policy."""


class AclCheck:
    """Check where hosts, ports and protocols match in a NAC policy.

    Attributes:
      pol_obj: policy.Policy object.
      pol: policy.Policy object.
      src: A string for the source address.
      dst: A string for the destination address.
      sport: A string for the source port.
      dport: A string for the destination port.
      proto: A string for the protocol.
      matches: A list of term-related matches.
      exact_matches: A list of exact matches.

    Returns:
      An AclCheck Object

    Raises:
      port.BadPortValue: An invalid source port is used
      port.BadPortRange: A port is outside of the acceptable range 0-65535
      AddressError: Incorrect ip address or format

    """

    @classmethod
    def FromPolicyDict(
        cls,
        policy_dict: policy_builder.PolicyDict,
        definitions: naming.Naming,
        src,
        dst,
        sport,
        dport,
        proto,
    ):
        """Construct an AclCheck object from a PolicyDict + Naming object."""
        policy_obj = policy.FromBuilder(policy_builder.PolicyBuilder(policy_dict, definitions))
        return cls(policy_obj, src, dst, sport, dport, proto)

    def __init__(
        self,
        pol: policy.Policy,
        src='any',
        dst='any',
        sport='any',
        dport='any',
        proto='any',
    ):
        self.pol_obj = pol
        self.proto = proto

        # validate source port
        if not sport or sport == 'any':
            self.sport = 'any'
        else:
            self.sport = port.Port(sport)

        # validate destination port
        if not dport or dport == 'any':
            self.dport = 'any'
        else:
            self.dport = port.Port(dport)

        # validate source address
        if not src or src == 'any':
            self.src = 'any'
        else:
            try:
                self.src = nacaddr.IP(src)
            except ValueError:
                raise AddressError(f'bad source address: {src}\n')

        # validate destination address
        if not dst or dst == 'any':
            self.dst = 'any'
        else:
            try:
                self.dst = nacaddr.IP(dst)
            except ValueError:
                raise AddressError(f'bad destination address: {dst}\n')

        if not isinstance(self.pol_obj, (policy.Policy)):
            raise BadPolicyError('Policy object is not valid.')

        self.matches = []
        self.exact_matches = []
        for header, terms in self.pol_obj.filters:
            filtername = header.target[0].options[0]
            for term in terms:
                possible = []
                logging.debug('checking term: %s', term.name)
                if not self._AddrInside(self.src, term.source_address):
                    logging.debug('srcaddr does not match')
                    continue
                logging.debug('srcaddr matches: %s', self.src)
                if not self._AddrInside(self.dst, term.destination_address):
                    logging.debug('dstaddr does not match')
                    continue
                logging.debug('dstaddr matches: %s', self.dst)
                if (
                    self.sport != 'any'
                    and term.source_port
                    and not self._PortInside(self.sport, term.source_port)
                ):
                    logging.debug('sport does not match')
                    continue
                logging.debug('sport matches: %s', self.sport)
                if (
                    self.dport != 'any'
                    and term.destination_port
                    and not self._PortInside(self.dport, term.destination_port)
                ):
                    logging.debug('dport does not match')
                    continue
                logging.debug('dport matches: %s', self.dport)
                if self.proto != 'any' and term.protocol and self.proto not in term.protocol:
                    logging.debug('proto does not match')
                    continue
                logging.debug('proto matches: %s', self.proto)
                if term.protocol_except and self.proto in term.protocol_except:
                    logging.debug('protocol excepted by term, no match.')
                    continue
                logging.debug('proto not excepted: %s', self.proto)
                if not term.action:  # avoid any verbatim
                    logging.debug('term had no action (verbatim?), no match.')
                    continue
                logging.debug('term has an action')
                possible = self._PossibleMatch(term)
                self.matches.append(Match(filtername, term.name, possible, term.action, term.qos))
                if possible:
                    logging.debug('term has options: %s, not treating as exact match', possible)
                    continue

                # if we get here then we have a match, and if the action isn't next and
                # there are no possibles, then this is a "definite" match and we needn't
                # look for any further matches (i.e. later terms may match, but since
                # we'll never get there we shouldn't report them)
                if 'next' not in term.action:
                    self.exact_matches.append(
                        Match(filtername, term.name, [], term.action, term.qos)
                    )
                    break

    def Matches(self):
        """Return list of matched terms."""
        return self.matches

    def ExactMatches(self):
        """Return matched terms, but not terms with possibles or action next."""
        return self.exact_matches

    def ActionMatch(self, action='any'):
        """Return list of matched terms with specified actions."""
        match_list = []
        for match in self.matches:
            if match.action:
                if not match.possibles:
                    if action == 'any' or action in match.action:
                        match_list.append(match)
        return match_list

    def DescribeMatches(self):
        """Provide sentence descriptions of matches.

        Returns:
          ret_str: text sentences describing matches
        """
        ret_str = []
        for match in self.matches:
            text = str(match)
            ret_str.append(text)
        return '\n'.join(ret_str)

    def __str__(self):
        text = []
        summary = self.Summarize()
        for filter, terms in summary.items():
            text.append(f"{' ' * 2}filter: {filter}")
            for matches in terms.values():
                text.append(matches['message'])
        return '\n'.join(text)

    def Summarize(self):
        summary = defaultdict(lambda: defaultdict(dict))
        for match in self.matches:
            summary[match.filter][match.term]["possibles"] = match.possibles
            text = []
            if match.possibles:
                text.append(f"{' ' * 10}term: {match.term} (possible match)")
                text.append(f"{' ' * 16}{match.action} if {match.possibles}")
            else:
                text.append(f"{' ' * 10}term: {match.term}")
                text.append(f"{' ' * 16}{match.action}")
            summary[match.filter][match.term]["message"] = '\n'.join(text)
        return summary

    def _PossibleMatch(self, term):
        """Ignore some options and keywords that are edge cases.

        Args:
          term: term object to examine for edge-cases

        Returns:
          ret_str: a list of reasons this term may possible match
        """
        ret_str = []
        if 'first-fragment' in term.option:
            ret_str.append('first-frag')
        if term.fragment_offset:
            ret_str.append('frag-offset')
        if term.packet_length:
            ret_str.append('packet-length')
        if 'established' in term.option:
            ret_str.append('est')
        if 'tcp-established' in term.option and 'tcp' in term.protocol:
            ret_str.append('tcp-est')
        return ret_str

    def _AddrInside(self, addr, addresses):
        """Check if address is matched in another address or group of addresses.

        Args:
          addr: An ipaddr network or host address or text 'any'
          addresses: A list of ipaddr network or host addresses

        Returns:
          bool: True of false
        """
        if addr == 'any':
            return True  # always true if we match for any addr
        if not addresses:
            return True  # always true if term has nothing to match
        for ip in addresses:
            # ipaddr can incorrectly report ipv4 as contained with ipv6 addrs
            if addr.subnet_of(ip):
                return True
        return False

    def _PortInside(self, myport, port_list):
        """Check if port matches in a port or group of ports.

        Args:
          myport: port number
          port_list: list of ports

        Returns:
          bool: True of false
        """
        if myport == 'any':
            return True
        if [x for x in port_list if x[0] <= myport <= x[1]]:
            return True
        return False


class Match:
    """A matching term and its associate values."""

    def __init__(self, filtername, term, possibles, action, qos=None):
        self.filter = filtername
        self.term = term
        self.possibles = possibles
        self.action = action[0]
        self.qos = qos

    def __str__(self):
        text = ''
        if self.possibles:
            text += f"possible {self.action}"
        else:
            text += self.action
        text += f" in term {self.term} of filter {self.filter}"
        if self.possibles:
            text += f" with factors: {', '.join(self.possibles)!s}"
        return text


def main():
    pass


if __name__ == '__main__':
    main()



"""YAML front-end. Loads a Policy model from a .yaml file."""

import pathlib
import typing
from typing import Optional, Union

import yaml
from absl import logging
from yaml.error import YAMLError

from aerleon.lib import policy
from aerleon.lib.policy import BadIncludePath, Policy, _SubpathOf
from aerleon.lib.policy_builder import PolicyBuilder, PolicyDict, TermsList
from aerleon.lib.yaml_loader import SpanSafeYamlLoader

if typing.TYPE_CHECKING:
    from aerleon.lib.naming import Naming


MAX_INCLUDE_DEPTH = 5


class PolicyTypeError(Exception):
    """Invalid policy."""


class ExcessiveRecursionError(Exception):
    """Include depth limit exceeded."""


# Consider making this span-oriented
# (file > line > (start_ch, end_ch))
class UserMessage:
    """A user-facing error message encountered during file processing.

    Users can be shown:
    * An error message only (user_message.message).
    * An error message with file / line / include stack (user_message.__repr__()).

    Attributes:
        message: The error message.
        filename: The name of the file in which this error or message originated.
        line: The line where this error or message originated.
        include_chain: If the error or message originated while processing an included
            file, include_chain will list the include file chain as a list of file/line tuples.
            The top-level file should be the first item in the list.
    """

    message: str
    filename: str
    line: int | None
    include_chain: "list[tuple[str, int]] | None"

    def __init__(
        self,
        message: str,
        *,
        filename: str,
        line: int | None = None,
        include_chain: "list[tuple[str, int]] | None" = None,
    ) -> None:
        self.message = message
        self.filename = filename
        self.line = line
        self.include_chain = include_chain

    def __str__(self) -> str:
        """Display user-facing error message with include chain (if present).

        e.g.
        Excessive recursion: include depth limit of 5 reached. File=include_1.yaml, Line=3.
        Include stack:
        > File='policy_with_include.yaml', Line=11 (Top Level)
        > File='include_1.yaml', Line=3
        > File='include_1.yaml', Line=3
        > File='include_1.yaml', Line=3
        > File='include_1.yaml', Line=3
        > File='include_1.yaml', Line=3
        """  # noqa: E501
        error_context = f"{self.message} File={self.filename}"
        if self.line is not None:
            error_context += f", Line={self.line}"
        error_context += "."
        if self.include_chain is not None and len(self.include_chain) > 1:
            error_context += "\nInclude stack:"
            for i, (File, Line) in enumerate(self.include_chain):
                error_context += f"\n> File='{File}', Line={Line}"
                if i == 0:
                    error_context += " (Top Level)"
        return error_context

    def __repr__(self):
        return f"UserMessage(\"{str(self)}\")"

    @classmethod
    def fromValueError(cls, error: ValueError, *, filename, line=None, include_chain=None):
        """Create a UserMessage from a ValueError."""
        return cls(str(error), filename=filename, line=line, include_chain=include_chain)


def ParseFile(
    filename, base_dir='', definitions: 'Naming | None' = None, optimize=False, shade_check=False
):
    """Load a policy yaml file and return a Policy data model.

    Args:
        filename: Policy file path. Any output configs will share
            the same file name (except the file extension).
        naming: Naming database (see Naming class). Resolves network
            names to networks or lists of networks.
        optimize: bool - Whether to summarize networks and services.
        shade_check: bool - Whether to raise an exception when a term is shaded.

    Raises:
        PolicyTypeError: The policy file provided is not valid.
    """
    with open(pathlib.Path(base_dir).joinpath(filename)) as file:
        try:
            policy_dict = yaml.load(file, Loader=SpanSafeYamlLoader(filename=filename))
        except YAMLError as yaml_error:
            raise PolicyTypeError(
                UserMessage("Unable to read file as YAML.", filename=filename)
            ) from yaml_error
    processor = YAMLPolicyPreprocessor(base_dir)
    policy_dict = processor(filename, policy_dict)
    if not policy_dict:
        return
    if not policy_dict['filters']:
        return
    return policy.FromBuilder(PolicyBuilder(policy_dict, definitions, optimize, shade_check))


def ParsePolicy(
    file: str,
    *,
    filename,
    base_dir='',
    definitions: 'Naming | None' = None,
    optimize=False,
    shade_check=False,
) -> Optional[Policy]:
    """Load a policy yaml file (provided as a string) and return a Policy data model.

    Note that "filename" must still be provided. The input filename is used to
    determine the output filename.

    Args:
        file: The contents of the policy file.
        filename: Any output configs base their file name on this value (except the file extension).
        naming: Naming database (see Naming class). Resolves network
            names to networks or lists of networks.
        optimize: bool - Whether to summarize networks and services.
        shade_check: bool - Whether to raise an exception when a term is shaded.

    Raises:
        PolicyTypeError: The policy file provided is not valid.
    """
    try:
        policy_dict = yaml.load(file, Loader=SpanSafeYamlLoader(filename=filename))
    except YAMLError as yaml_error:
        raise PolicyTypeError(
            UserMessage("Unable to read file as YAML.", filename=filename)
        ) from yaml_error

    processor = YAMLPolicyPreprocessor(base_dir)
    policy_dict = processor(filename, policy_dict)
    if not policy_dict:
        return
    if not policy_dict['filters']:
        return
    return policy.FromBuilder(PolicyBuilder(policy_dict, definitions, optimize, shade_check))


def suffix_is_yaml(filename):
    return filename[-5:] == '.yaml' or filename[-4:] == '.yml'


class YAMLPolicyPreprocessor:
    """Processes a policy dictionary, handling includes and performing validation."""

    def __init__(self, base_dir: str):
        """
        Args:
            base_dir: The base directory for resolving include paths.
        """
        self.base_dir = base_dir

    def __call__(self, filename: str, policy_dict: Optional[PolicyDict]) -> Optional[PolicyDict]:
        """Process includes and validate the file data as a PolicyDict.

        Args:
            filename: The name of the policy file.
            policy_dict: The parsed YAML policy data.

        Returns:
            The processed policy dictionary with includes expanded.
        """
        debug_stack = []
        return self._preprocess_inner(
            MAX_INCLUDE_DEPTH,
            debug_stack,
            filename=filename,
            policy_dict=policy_dict,
        )

    def _preprocess_inner(
        self, depth: int, debug_stack: list, filename: str, policy_dict: Optional[PolicyDict]
    ) -> Optional[PolicyDict]:
        # Empty files are ignored with a warning
        if policy_dict is None or not policy_dict:
            logging.warning(UserMessage("Ignoring empty policy file.", filename=filename))
            return

        # Malformed policy files should generate a PolicyTypeError (unless this is an include file)
        if 'filters' in policy_dict and isinstance(policy_dict['filters'], list):
            pass  # Normal case.
        elif (
            depth < MAX_INCLUDE_DEPTH
            and 'filters_include_only' in policy_dict
            and isinstance(policy_dict['filters_include_only'], list)
        ):
            # Policy files with filters_include_only: are ignored by ParsePolicy but can be included.
            policy_dict['filters'] = policy_dict['filters_include_only']
            del policy_dict['filters_include_only']
        elif 'terms' in policy_dict or 'filters_include_only' in policy_dict:
            # We are looking at an include file outside of an include and should quietly ignore it.
            return
        else:
            raise PolicyTypeError(
                UserMessage(
                    "Policy file must contain one or more filter sections.", filename=filename
                )
            )

        found_filters = []

        for filter_item in policy_dict['filters']:
            # Malformed filters should generate a PolicyTypeError
            if not isinstance(filter_item, dict):
                raise PolicyTypeError(UserMessage("Filter must be a mapping.", filename=filename))

            def expand_filter(filter_to_expand):
                stack = debug_stack.copy()
                stack.append((filter_to_expand['__filename__'], filter_to_expand['__line__']))

                if depth <= 0:
                    raise ExcessiveRecursionError(
                        UserMessage(
                            f"Excessive recursion: include depth limit of {MAX_INCLUDE_DEPTH} reached.",  # noqa: E501
                            filename=filter_to_expand['__filename__'],
                            line=filter_to_expand['__line__'],
                            include_chain=stack,
                        )
                    )
                try:
                    include_data, include_path = self._load_include_file(
                        filter_to_expand['include'], stack
                    )
                except ValueError as value_error:
                    raise PolicyTypeError(
                        UserMessage.fromValueError(
                            value_error,
                            filename=filter_to_expand['__filename__'],
                            line=filter_to_expand['__line__'],
                            include_chain=stack,
                        )
                    ) from value_error
                except YAMLError as yaml_error:
                    raise PolicyTypeError(
                        UserMessage(
                            "Unable to read file as YAML.",
                            filename=str(
                                pathlib.Path(self.base_dir).joinpath(filter_to_expand['include'])
                            ),
                            include_chain=stack,
                        )
                    ) from yaml_error

                data = self._preprocess_inner(
                    depth - 1,
                    stack,
                    filename=include_path.name,
                    policy_dict=include_data,
                )
                if not (data and data['filters']):
                    logging.warning(
                        UserMessage(
                            "Ignoring empty policy include source.",
                            filename=str(include_path),
                            include_chain=stack,
                        )
                    )
                    return
                found_filters.extend(data['filters'])

            if 'include' in filter_item:
                # This is an include directive
                expand_filter(filter_item)
                continue

            if 'header' not in filter_item or not isinstance(filter_item['header'], dict):
                raise PolicyTypeError(
                    UserMessage(
                        "Filter must contain a header section.",
                        filename=filename,
                        line=filter_item['__line__'],
                    )
                )

            if (
                'terms' not in filter_item
                or not filter_item['terms']
                or not isinstance(filter_item['terms'], list)
            ):
                raise PolicyTypeError(
                    UserMessage(
                        "Filter must contain a terms section.",
                        filename=filename,
                        line=filter_item['__line__'],
                    )
                )

            header = filter_item['header']
            if 'targets' not in header or (
                header['targets'] is not None and not isinstance(header['targets'], dict)
            ):
                raise PolicyTypeError(
                    UserMessage(
                        "Filter header must contain a targets section.",
                        filename=filename,
                        line=header['__line__'],
                    )
                )

            # Filters with an empty target list can be ignored with a warning
            elif not header['targets']:
                raise PolicyTypeError(
                    UserMessage(
                        "Filter header cannot be empty.",
                        filename=filename,
                        line=filter_item['__line__'],
                    )
                )

            found_terms = []

            def process_terms(term_depth, stack, term_items):
                for term_item in term_items:
                    if 'include' not in term_item:
                        found_terms.append(term_item)
                        continue
                    new_stack = stack.copy()
                    new_stack.append((term_item['__filename__'], term_item['__line__']))
                    if term_depth <= 0:
                        raise ExcessiveRecursionError(
                            UserMessage(
                                f"Excessive recursion: include depth limit of {MAX_INCLUDE_DEPTH} reached.",  # noqa: E501
                                filename=term_item['__filename__'],
                                line=term_item['__line__'],
                                include_chain=new_stack,
                            )
                        )
                    try:
                        include_data, include_path = self._load_include_file(
                            term_item['include'], new_stack
                        )
                    except ValueError as value_error:
                        raise PolicyTypeError(
                            UserMessage.fromValueError(
                                value_error,
                                filename=term_item['__filename__'],
                                line=term_item['__line__'],
                                include_chain=new_stack,
                            )
                        ) from value_error
                    except YAMLError as yaml_error:
                        raise PolicyTypeError(
                            UserMessage(
                                "Unable to read file as YAML.",
                                filename=str(
                                    pathlib.Path(self.base_dir).joinpath(term_item['include'])
                                ),
                                include_chain=new_stack,
                            )
                        ) from yaml_error
                    if (
                        not include_data
                        or 'terms' not in include_data
                        or not include_data['terms']
                    ):
                        logging.warning(
                            UserMessage(
                                "Ignoring empty policy include source.",
                                filename=str(include_path),
                                include_chain=new_stack,
                            )
                        )
                        continue
                    process_terms(term_depth - 1, new_stack, include_data['terms'])

            process_terms(MAX_INCLUDE_DEPTH, [], filter_item['terms'])

            if not found_terms:
                logging.warning(
                    UserMessage(
                        "Ignoring filter with zero terms.",
                        filename=filename,
                        line=filter_item['__line__'],
                    )
                )
                continue

            for term_item in found_terms:
                if 'name' not in term_item or len(term_item['name'].strip()) == 0:
                    raise PolicyTypeError(
                        UserMessage(
                            "Term must have a name.",
                            filename=term_item['__filename__'],
                            line=term_item['__line__'],
                        )
                    )

            filter_item['terms'] = found_terms
            found_filters.append(filter_item)

        policy_dict['filters'] = found_filters

        def StripDebuggingData(data):
            if isinstance(data, list):
                for item in data:
                    if isinstance(item, list) or isinstance(item, dict):
                        StripDebuggingData(item)
            elif isinstance(data, dict):
                data.pop('__line__', None)
                data.pop('__filename__', None)
                for item in data.values():
                    if isinstance(item, list) or isinstance(item, dict):
                        StripDebuggingData(item)

        StripDebuggingData(policy_dict)

        return policy_dict

    def _load_include_file(
        self, relative_path: str, stack: list
    ) -> tuple[Optional[PolicyDict], Union[str, pathlib.Path]]:
        """Load, parse, and validate an include file path."""
        if not suffix_is_yaml(relative_path):
            raise ValueError(
                f'Policy include source {relative_path} must end in ".yaml" or ".yml".'
            )
        include_path = pathlib.Path(self.base_dir).joinpath(relative_path)
        if not _SubpathOf(self.base_dir, include_path):
            raise BadIncludePath(
                f"Include file cannot be loaded from outside the base directory. File={include_path} base_directory={self.base_dir}"
            )

        with open(include_path) as include_file:
            include_data = yaml.load(
                include_file.read(), Loader=SpanSafeYamlLoader(filename=str(include_path))
            )
        return include_data, include_path


class GenerateAPIPolicyPreprocessor(YAMLPolicyPreprocessor):
    """A YAMLPolicyPreprocessor that sources includes from a dictionary."""

    def __init__(self, includes: dict[str, TermsList]):
        """
        Args:
            includes: A read-only mapping from include name to file_dict.
        """
        super().__init__('')
        self.includes = includes

    def _load_include_file(
        self, relative_path: str, stack: list
    ) -> tuple[Optional[TermsList], Union[str, pathlib.Path]]:
        """Override to load includes from the self.includes dictionary."""
        return self.includes.get(relative_path), relative_path



# Copyright 2010 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Iptables generator."""

import re
from string import Template  # pylint: disable=g-importing-member
from typing import Optional, Union

from absl import logging

from aerleon.lib import aclgenerator, nacaddr
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Policy, Term


class Term(aclgenerator.Term):
    """Generate Iptables policy terms."""

    # Validate that term does not contain any fields we do not
    # support.  This prevents us from thinking that our output is
    # correct in cases where we've omitted fields from term.
    _PLATFORM = 'iptables'
    _POSTJUMP_FORMAT = None
    _PREJUMP_FORMAT = Template('-A $filter -j $term')
    _TERM_FORMAT = Template('-N $term')
    _COMMENT_FORMAT = Template('-A $term -m comment --comment "$comment"')
    _FILTER_TOP_FORMAT = Template('-A $term')
    _LOG_FORMAT = Template('-j LOG --log-prefix $term')
    _PROTO_TABLE = {
        'icmpv6': '-p ipv6-icmp',
        'icmp': '-p icmp',
        'tcp': '-p tcp',
        'udp': '-p udp',
        'all': '-p all',
        'esp': '-p esp',
        'ah': '-p ah',
        'gre': '-p gre',
    }
    _TCP_FLAGS_TABLE = {
        'syn': 'SYN',
        'ack': 'ACK',
        'fin': 'FIN',
        'rst': 'RST',
        'urg': 'URG',
        'psh': 'PSH',
        'all': 'ALL',
        'none': 'NONE',
    }
    _KNOWN_OPTIONS_MATCHERS = {
        # '! -f' also matches non-fragmented packets.
        'first-fragment': '-m u32 --u32 4&0x3FFF=0x2000',
        'initial': '--syn',
        'tcp-initial': '--syn',
        'sample': '',
    }

    def __init__(
        self,
        term: Term,
        filter_name: str,
        trackstate: bool,
        filter_action: Optional[str],
        af: str = 'inet',
        verbose: bool = True,
        chained_terms: bool = True,
    ) -> None:
        """Setup a new term.

        Args:
          term: A policy.Term object to represent in iptables.
          filter_name: The name of the filter chan to attach the term to.
          trackstate: Specifies if conntrack should be used for new connections.
          filter_action: The default action of the filter.
          af: Which address family ('inet' or 'inet6') to apply the term to.
          verbose: boolean if comments should be printed

        Raises:
          UnsupportedFilterError: Filter is not supported.
        """
        super().__init__(term)
        self._action_table = {
            'accept': '-j ACCEPT',
            'deny': '-j DROP',
            'reject': '-j REJECT --reject-with icmp-host-prohibited',
            'reject-with-tcp-rst': '-j REJECT --reject-with tcp-reset',
            'next': '-j RETURN',
        }
        self.trackstate = trackstate
        self.term = term  # term object
        self.filter = filter_name  # actual name of filter
        self.default_action = filter_action
        self.options = []
        self.af = af
        self.verbose = verbose
        self.chained_terms = chained_terms
        if af == 'inet6':
            self._all_ips = nacaddr.IPv6('::/0')
            self._action_table['reject'] = '-j REJECT --reject-with ' 'icmp6-adm-prohibited'
        else:
            self._all_ips = nacaddr.IPv4('0.0.0.0/0')
            self._action_table['reject'] = '-j REJECT --reject-with ' 'icmp-host-prohibited'
        if self.chained_terms:
            self.term_name = f'{self.filter[:1]}_{self.term.name}'
        else:
            self.term_name = self.filter

    def __str__(self) -> str:
        ret_str = []

        # Don't render icmpv6 protocol terms under inet, or icmp under inet6
        if (self.af == 'inet6' and 'icmp' in self.term.protocol) or (
            self.af == 'inet' and 'icmpv6' in self.term.protocol
        ):
            logging.warning(
                self.NO_AF_LOG_PROTO.substitute(
                    term=self.term.name, proto=', '.join(self.term.protocol), af=self.af
                )
            )
            return ''

        # Term verbatim output - this will skip over most normal term
        # creation code by returning early. Warnings provided in policy.py
        if self.term.verbatim:
            for next_verbatim in self.term.verbatim:
                if next_verbatim[0] == self._PLATFORM:
                    ret_str.append(str(next_verbatim[1]))
            return '\n'.join(ret_str)

        # Create a new term
        self._SetDefaultAction()

        if self.chained_terms and self._TERM_FORMAT:
            ret_str.append(self._TERM_FORMAT.substitute(term=self.term_name))
        if self.chained_terms and self._PREJUMP_FORMAT:
            ret_str.append(
                self._PREJUMP_FORMAT.substitute(filter=self.filter, term=self.term_name)
            )

        if self.verbose:
            if self.term.owner:
                self.term.comment.append(f'Owner: {self.term.owner}')
            # reformat long comments, if needed
            #
            # iptables allows individual comments up to 256 chars.
            # But our generator will limit a single comment line to < 120, using:
            # max = 119 - 27 (static chars in comment command) - [length of term name]
            comment_max_width = 92 - len(self.term_name)
            if comment_max_width < 40:
                comment_max_width = 40
            comments = aclgenerator.WrapWords(self.term.comment, comment_max_width)
            # append comments to output
            if comments and comments[0]:
                for line in comments:
                    if not line:
                        continue  # iptables-restore does not like 0-length comments.
                    # term comments
                    # Strip out quotes as iptables cant have nested quotes
                    ret_str.append(
                        self._COMMENT_FORMAT.substitute(
                            filter=self.filter,
                            term=self.term_name,
                            comment=str(line).replace('\"', ''),
                        )
                    )

        # Unsupported configuration; in the case of 'accept' or 'next', we
        # skip the rule.  In other cases, we blow up (raise an exception)
        # to ensure that this is not considered valid configuration.
        if self.term.source_prefix or self.term.destination_prefix:
            if str(self.term.action[0]) not in {'accept', 'next'}:
                raise UnsupportedFilterError(
                    '%s %s %s %s %s %s %s %s'
                    % (
                        '\nTerm',
                        self.term.name,
                        'has action',
                        str(self.term.action[0]),
                        'with source_prefix or destination_prefix,',
                        ' which is unsupported in',
                        self._PLATFORM,
                        'iptables output.',
                    )
                )
            return f'# skipped {self.term.name} due to source or destination prefix rule'

        # protocol
        if self.term.protocol:
            protocol = self.term.protocol
        else:
            protocol = ['all']
        if 'hopopt' in protocol and self.af == 'inet':
            logging.warning('Term %s is using hopopt in IPv4 context.', self.term_name)
            return ''

        (term_saddr, exclude_saddr, term_daddr, exclude_daddr) = self._CalculateAddresses(
            self.term.source_address,
            self.term.source_address_exclude,
            self.term.destination_address,
            self.term.destination_address_exclude,
        )
        if not term_saddr:
            logging.warning(
                self.NO_AF_LOG_ADDR.substitute(term=self.term.name, direction='source', af=self.af)
            )
            return ''
        if not term_daddr:
            logging.warning(
                self.NO_AF_LOG_ADDR.substitute(
                    term=self.term.name, direction='destination', af=self.af
                )
            )
            return ''

        # ports
        source_port = []
        destination_port = []
        if self.term.source_port:
            source_port = self.term.source_port
        if self.term.destination_port:
            destination_port = self.term.destination_port

        # icmp-types
        icmp_types = ['']
        if self.term.icmp_type:
            icmp_types = self.NormalizeIcmpTypes(self.term.icmp_type, protocol, self.af)

        source_interface = ''
        if self.term.source_interface:
            source_interface = self.term.source_interface

        destination_interface = ''
        if self.term.destination_interface:
            destination_interface = self.term.destination_interface

        log_hits = False
        if self.term.logging:
            # Iptables sends logs to hosts configured syslog
            log_hits = True

        # options
        tcp_flags = []
        tcp_track_options = []
        for next_opt in [str(x) for x in self.term.option]:
            #
            # Sanity checking and high-ports are added as appropriate in
            # pre-processing that is done in __str__ within class Iptables.
            # Option established will add destination port high-ports if protocol
            # contains only tcp, udp or both.  This is done earlier in class Iptables.
            #
            if (
                next_opt.find('established') == 0 or next_opt.find('tcp-established') == 0
            ) and 'ESTABLISHED' not in [x.strip() for x in self.options]:
                if next_opt.find('tcp-established') == 0 and protocol != ['tcp']:
                    raise TcpEstablishedError(
                        '%s %s %s'
                        % (
                            '\noption tcp-established can only be applied for proto tcp.',
                            '\nError in term:',
                            self.term.name,
                        )
                    )

                if self.trackstate:
                    # Use nf_conntrack to track state -- works with any proto
                    self.options.append('-m state --state ESTABLISHED,RELATED')
                elif protocol == ['tcp']:
                    # Simple established-only rule for TCP: Must have ACK field
                    # (SYN/ACK or subsequent ACK), or RST and no other flags.
                    tcp_track_options = [
                        (['ACK'], ['ACK']),
                        (['SYN', 'FIN', 'ACK', 'RST'], ['RST']),
                    ]

            # Iterate through flags table, and create list of tcp-flags to append
            for next_flag in self._TCP_FLAGS_TABLE:
                if next_opt.find(next_flag) == 0:
                    tcp_flags.append(self._TCP_FLAGS_TABLE.get(next_flag))
            if next_opt in self._KNOWN_OPTIONS_MATCHERS:
                self.options.append(self._KNOWN_OPTIONS_MATCHERS[next_opt])
        if self.term.packet_length:
            # Policy format is "#-#", but iptables format is "#:#"
            self.options.append(f"-m length --length {self.term.packet_length.replace('-', ':')}")
        if self.term.fragment_offset:
            self.options.append(
                f"-m u32 --u32 4&0x1FFF={self.term.fragment_offset.replace('-', ':')}"
            )
        icmp_code = ['']
        if self.term.icmp_code:
            icmp_code = self.term.icmp_code

        for saddr in exclude_saddr:
            ret_str.extend(
                self._FormatPart(
                    '',
                    saddr,
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    self._action_table.get('next'),
                )
            )
        for daddr in exclude_daddr:
            ret_str.extend(
                self._FormatPart(
                    '',
                    '',
                    '',
                    daddr,
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    '',
                    self._action_table.get('next'),
                )
            )

        for saddr in term_saddr:
            for daddr in term_daddr:
                for icmp in icmp_types:
                    for code in icmp_code:
                        for proto in protocol:
                            for tcp_matcher in tcp_track_options or (([], []),):
                                ret_str.extend(
                                    self._FormatPart(
                                        str(proto),
                                        saddr,
                                        source_port,
                                        daddr,
                                        destination_port,
                                        self.options,
                                        tcp_flags,
                                        icmp,
                                        code,
                                        tcp_matcher,
                                        source_interface,
                                        destination_interface,
                                        log_hits,
                                        self._action_table.get(str(self.term.action[0])),
                                    )
                                )

        if self._POSTJUMP_FORMAT and self.chained_terms:
            ret_str.append(
                self._POSTJUMP_FORMAT.substitute(filter=self.filter, term=self.term_name)
            )

        return '\n'.join(str(v) for v in ret_str if v)

    def _CalculateAddresses(
        self,
        term_saddr: list[Union[IPv4, IPv6]],
        exclude_saddr: list[Union[IPv4, IPv6]],
        term_daddr: list[Union[IPv4, IPv6]],
        exclude_daddr: list[Union[IPv4, IPv6]],
    ) -> Union[
        tuple[
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
        ],
        tuple[
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
        ],
        tuple[
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
        ],
        tuple[
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
            list[Union[IPv4, IPv6]],
        ],
    ]:
        """Calculate source and destination address list for a term.

        Args:
          term_saddr: source address list of the term
          exclude_saddr: source address exclude list of the term
          term_daddr: destination address list of the term
          exclude_daddr: destination address exclude list of the term

        Returns:
          tuple containing source address list, source exclude address list,
          destination address list, destination exclude address list in
          that order

        """
        # source address
        term_saddr_excluded = []
        if not term_saddr:
            term_saddr = [self._all_ips]
        if exclude_saddr:
            term_saddr_excluded.extend(nacaddr.ExcludeAddrs(term_saddr, exclude_saddr))

        # destination address
        term_daddr_excluded = []
        if not term_daddr:
            term_daddr = [self._all_ips]
        if exclude_daddr:
            term_daddr_excluded.extend(nacaddr.ExcludeAddrs(term_daddr, exclude_daddr))

        # Just to be safe, always have a result of at least 1 to avoid * by zero
        # returning incorrect results (10src*10dst=100, but 10src*0dst=0, not 10)
        bailout_count = (
            len(exclude_saddr)
            + len(exclude_daddr)
            + ((len(self.term.source_address) or 1) * (len(self.term.destination_address) or 1))
        )
        exclude_count = (len(term_saddr_excluded) or 1) * (len(term_daddr_excluded) or 1)

        # Use bailout jumps for excluded addresses if it results in fewer output
        # lines than nacaddr.ExcludeAddrs() method.
        if exclude_count < bailout_count:
            exclude_saddr = []
            exclude_daddr = []
            if term_saddr_excluded:
                term_saddr = term_saddr_excluded
            if term_daddr_excluded:
                term_daddr = term_daddr_excluded

        # With many sources and destinations, iptables needs to generate the
        # cartesian product of sources and destinations.  If there are no
        # exclude rules, this can instead be written as exclude [0/0 -
        # srcs], exclude [0/0 - dsts].
        v4_src_count = len([x for x in term_saddr if x.version == 4])
        v4_dst_count = len([x for x in term_daddr if x.version == 4])
        v6_src_count = len([x for x in term_saddr if x.version == 6])
        v6_dst_count = len([x for x in term_daddr if x.version == 6])
        num_pairs = v4_src_count * v4_dst_count + v6_src_count * v6_dst_count
        if num_pairs > 100:
            new_exclude_source = nacaddr.ExcludeAddrs([self._all_ips], term_saddr)
            new_exclude_dest = nacaddr.ExcludeAddrs([self._all_ips], term_daddr)
            # Invert the shortest list that does not already have exclude addresses
            if len(new_exclude_source) < len(new_exclude_dest) and not exclude_saddr:
                if len(new_exclude_source) + len(term_daddr) < num_pairs:
                    exclude_saddr = new_exclude_source
                    term_saddr = [self._all_ips]
            elif not exclude_daddr:
                if len(new_exclude_dest) + len(term_saddr) < num_pairs:
                    exclude_daddr = new_exclude_dest
                    term_daddr = [self._all_ips]
        term_saddr = [x for x in term_saddr if x.version == self.AF_MAP[self.af]]
        exclude_saddr = [x for x in exclude_saddr if x.version == self.AF_MAP[self.af]]
        term_daddr = [x for x in term_daddr if x.version == self.AF_MAP[self.af]]
        exclude_daddr = [x for x in exclude_daddr if x.version == self.AF_MAP[self.af]]
        return (term_saddr, exclude_saddr, term_daddr, exclude_daddr)

    def _FormatPart(
        self,
        protocol: str,
        saddr: Union[IPv6, IPv4, str],
        sport: Union[list[tuple[int, int]], str],
        daddr: Union[IPv6, IPv4, str],
        dport: Union[list[tuple[int, int]], str],
        options: Union[str, list[str]],
        tcp_flags: Union[str, list[str]],
        icmp_type: Union[int, str],
        code: Union[int, str],
        track_flags: Union[tuple[list[str], list[str]], str],
        sint: str,
        dint: str,
        log_hits: Union[str, bool],
        action: str,
    ) -> list[str]:
        """Compose one iteration of the term parts into a string.

        Args:
          protocol: The network protocol
          saddr: Source IP address
          sport: Source port numbers
          daddr: Destination IP address
          dport: Destination port numbers
          options: Optional arguments to append to our rule
          tcp_flags: Which tcp_flag arguments, if any, should be appended
          icmp_type: What icmp protocol to allow, if any
          code: ICMP code allowed, if any
          track_flags: A tuple of ([check-flags], [set-flags]) arguments to tcp-flag
          sint: Optional source interface
          dint: Optional destination interface
          log_hits: Boolean, to log matches or not
          action: What should happen if this rule matches
        Returns:
          rval:  A single iptables argument line
        """
        src, dst = self._GenerateAddressStatement(saddr, daddr)

        filter_top = self._FILTER_TOP_FORMAT.substitute(filter=self.filter, term=self.term_name)

        source_int = ''
        if sint:
            source_int = f'-i {sint}'

        destination_int = ''
        if dint:
            destination_int = f'-o {dint}'

        log_jump = ''
        if log_hits:
            log_jump = self._LOG_FORMAT.substitute(term=self.term.name)
            if self.term.log_limit:
                log_jump = '-m --limit {}/{} {}'.format(
                    self.term.log_limit[0], self.term.log_limit[1], log_jump
                )

        if not options:
            options = []

        proto = self._PROTO_TABLE.get(str(protocol))
        # Don't drop protocol if we don't recognize it
        if protocol and not proto:
            proto = f'-p {protocol!s}'

        if protocol == 'hopopt':
            proto = ''
            # Select 4 bytes at offset 0x3, mask out all but
            # last byte. That produces byte at position 7,
            # Next Header. Compare to 0x0, Hop By Hop
            options.append('-m u32 --u32 "0x3&0xff=0x0"')
        if protocol == 'fragment':
            proto = ''
            # Ditto, but compare to 0x2c, 44, Fragment
            options.append('-m u32 --u32 "0x3&0xff=0x2c"')

        # set conntrack state to NEW, unless policy requested "nostate"
        # for icmpv6 use trackstate only for type 128 and 139, as this is the
        # only supported types in nf_conntrack_proto_icmpv6
        if self.trackstate and (protocol != 'icmpv6' or icmp_type in [128, 139]):
            already_stateful = False
            # we will add new stateful arguments only if none already exist, such
            # as from "option:: established"
            for option in options:
                if 'state' in option:
                    already_stateful = True
            if not already_stateful:
                if 'ACCEPT' in action:
                    # We have to permit established/related since a policy may not
                    # have an existing blank permit for established/related, which
                    # may be more efficient, but slightly less secure.
                    options.append('-m state --state NEW,ESTABLISHED,RELATED')

        if tcp_flags or (track_flags and track_flags[0]):
            check_fields = ','.join(sorted(set(tcp_flags + track_flags[0])))
            set_fields = ','.join(sorted(set(tcp_flags + track_flags[1])))
            flags = f'--tcp-flags {check_fields} {set_fields}'
        else:
            flags = ''

        icmp_type = str(icmp_type)
        if not icmp_type:
            icmp = ''
        elif str(protocol) == 'icmpv6':
            icmp = f'-m icmp6 --icmpv6-type {icmp_type}'
        else:
            icmp = f'--icmp-type {icmp_type}'
        if code:
            icmp += r'/%d' % code

        # format tcp and udp ports
        sports = dports = ['']
        if sport:
            sports = self._GeneratePortStatement(sport, source=True)
        if dport:
            dports = self._GeneratePortStatement(dport, dest=True)

        ret_lines = []
        for sport in sports:
            for dport in dports:
                rval = [filter_top]
                if re.search('multiport', sport) and not re.search('multiport', dport):
                    # Due to bug in iptables, use of multiport module before a single
                    # port specification will result in multiport trying to consume it.
                    # this is a little hack to ensure single ports are listed before
                    # any multiport specification.
                    dport, sport = sport, dport
                if str(protocol) == 'icmpv6':
                    # Due to a bug in ip6tables, iptables-save returns icmpv6 matches in
                    # order (address spec) (icmpv6 spec). Fake this using options
                    # datastructure.
                    options.extend((proto, icmp))
                    proto = ''
                    icmp = ''
                if str(self.af) == 'inet6' and str(protocol) == 'all' and 'REJECT' in str(action):
                    # Due to a bug in ip6tables, when -p all and -j REJECT, proto
                    # is being eaten
                    proto = ''
                for value in (
                    proto,
                    flags,
                    sport,
                    dport,
                    icmp,
                    src,
                    dst,
                    ' '.join(options),
                    source_int,
                    destination_int,
                ):
                    if value:
                        rval.append(str(value))
                if log_jump:
                    # -j LOG
                    ret_lines.append(' '.join(rval + [log_jump]))
                # -j ACTION
                ret_lines.append(' '.join(rval + [action]))
        return ret_lines

    def _GenerateAddressStatement(
        self, saddr: Union[IPv6, IPv4], daddr: Union[IPv6, IPv4]
    ) -> tuple[str, str]:
        """Return the address section of an individual iptables rule.

        Args:
          saddr: source address of the rule
          daddr: destination address of the rule

        Returns:
          tuple containing source and destination address statement, in
          that order

        """
        src = ''
        dst = ''
        if not saddr or saddr == self._all_ips:
            src = ''
        else:
            src = '-s %s/%d' % (saddr.network_address, saddr.prefixlen)
        if not daddr or daddr == self._all_ips:
            dst = ''
        else:
            dst = '-d %s/%d' % (daddr.network_address, daddr.prefixlen)
        return (src, dst)

    def _GeneratePortStatement(
        self, ports: list[tuple[int, int]], source: bool = False, dest: bool = False
    ) -> list[str]:
        """Return the 'port' section of an individual iptables rule.

        Args:
          ports: list of ports or port ranges (pairs)
          source: (bool) generate a source port rule
          dest: (bool) generate a dest port rule

        Returns:
          list holding the 'port' sections of an iptables rule.

        Raises:
          BadPortsError: if too many ports are passed in, or if both 'source'
                            and 'dest' are true.
          NotImplementedError: if both 'source' and 'dest' are true.
        """
        if not ports:
            return ''

        direction = ''  # default: no direction / '--port'.  As yet, unused.
        if source and dest:
            raise BadPortsError('_GeneratePortStatement called ambiguously.')
        elif source:
            direction = 's'  # source port / '--sport'
        elif dest:
            direction = 'd'  # dest port / '--dport'
        else:
            raise NotImplementedError('--port support not yet implemented.')

        # Normalize ports and get accurate port count.
        # iptables multiport module limits to 15, but we use 14 to ensure a range
        # doesn't tip us over the limit
        max_ports = 14
        norm_ports = []
        portstrings = []
        count = 0
        for port in ports:
            if port[0] == port[1]:
                norm_ports.append(str(port[0]))
                count += 1
            else:
                norm_ports.append('%d:%d' % (port[0], port[1]))
                count += 2
            if count >= max_ports:
                count = 0
                portstrings.append(f"-m multiport --{direction}ports {','.join(norm_ports)}")
                norm_ports = []
        if norm_ports:
            if len(norm_ports) == 1:
                portstrings.append(f'--{direction}port {norm_ports[0]}')
            else:
                portstrings.append(f"-m multiport --{direction}ports {','.join(norm_ports)}")
        return portstrings

    def _SetDefaultAction(self) -> None:
        """If term does not specify action, use filter default action."""
        if not self.term.action:
            self.term.action[0].value = self.default_action


class Iptables(aclgenerator.ACLGenerator):
    """Generates filters and terms from provided policy object."""

    _PLATFORM = 'iptables'
    _DEFAULT_PROTOCOL = 'all'
    SUFFIX = ''
    _RENDER_PREFIX = None
    _RENDER_SUFFIX = None
    _DEFAULTACTION_FORMAT = '-P %s %s'
    _DEFAULTACTION_FORMAT_CUSTOM_CHAIN = '-N %s'
    _DEFAULT_ACTION = 'DROP'
    _TERM = Term
    _TERM_MAX_LENGTH = 24
    _GOOD_FILTERS = ['INPUT', 'OUTPUT', 'FORWARD']
    _GOOD_OPTIONS = ['nostate', 'abbreviateterms', 'truncateterms', 'noverbose', 'nochainedterms']

    def __init__(self, pol: Policy, exp_info: int) -> None:
        self.iptables_policies = []
        super().__init__(pol, exp_info)

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {
            'counter',
            'destination_interface',
            'destination_prefix',
            'fragment_offset',
            'icmp_code',
            'logging',
            'log_limit',
            'owner',
            'packet_length',
            'routing_instance',
            'source_interface',
            'source_prefix',
        }

        supported_sub_tokens.update(
            {
                'option': {
                    'established',
                    'first-fragment',
                    'initial',
                    'sample',
                    'tcp-established',
                    'tcp-initial',
                    'syn',
                    'ack',
                    'fin',
                    'rst',
                    'urg',
                    'psh',
                    'all',
                    'none',
                },
            }
        )
        return supported_tokens, supported_sub_tokens

    def _WarnIfCustomTarget(self, target: str) -> None:
        """Emit a warning if a policy's default target is not a built-in chain."""
        if target not in self._GOOD_FILTERS:
            logging.warning(
                'Filter is generating a non-standard chain that will not '
                'apply to traffic unless linked from INPUT, OUTPUT or '
                'FORWARD filters. New chain name is: %s',
                target,
            )

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        """Translate a policy from objects into strings."""
        default_action = None
        good_default_actions = ['ACCEPT', 'DROP']
        good_afs = ['inet', 'inet6']
        chained_terms = True
        all_protocols_stateful = True
        self.verbose = True

        for header, terms in pol.filters:
            filter_type = None
            self.filter_options = header.FilterOptions(self._PLATFORM)[1:]
            filter_name = header.FilterName(self._PLATFORM)

            self._WarnIfCustomTarget(filter_name)

            # ensure all options after the filter name are expected
            for opt in self.filter_options:
                if opt not in good_default_actions + good_afs + self._GOOD_OPTIONS:
                    raise UnsupportedTargetOptionError(
                        '%s %s %s %s'
                        % (
                            '\nUnsupported option found in',
                            self._PLATFORM,
                            'target definition:',
                            opt,
                        )
                    )

            # disable stateful?
            if 'nostate' in self.filter_options:
                all_protocols_stateful = False
            if 'noverbose' in self.filter_options:
                self.verbose = False
            if 'nochainedterms' in self.filter_options:
                chained_terms = False

            # Check for matching af
            for address_family in good_afs:
                if address_family in self.filter_options:
                    # should not specify more than one AF in options
                    if filter_type is not None:
                        raise UnsupportedFilterError(
                            '%s %s %s %s'
                            % (
                                '\nMay only specify one of',
                                good_afs,
                                'in filter options:',
                                self.filter_options,
                            )
                        )
                    filter_type = address_family
            if filter_type is None:
                filter_type = 'inet'

            if self._PLATFORM == 'iptables' and filter_name == 'FORWARD':
                default_action = 'DROP'

            # does this policy override the default filter actions?
            for next_target in header.target:
                if next_target.platform == self._PLATFORM:
                    if len(next_target.options) > 1:
                        for arg in next_target.options:
                            if arg in good_default_actions:
                                default_action = arg
            if default_action and default_action not in good_default_actions:
                raise UnsupportedDefaultActionError(
                    '%s %s %s %s %s'
                    % (
                        '\nOnly',
                        ', '.join(good_default_actions),
                        'default filter action allowed;',
                        default_action,
                        'used.',
                    )
                )

            # add the terms
            new_terms = []
            term_names = set()
            for term in terms:
                term.name = self.FixTermLength(
                    term.name,
                    'abbreviateterms' in self.filter_options,
                    'truncateterms' in self.filter_options,
                )
                if term.name in term_names:
                    raise aclgenerator.DuplicateTermError(
                        f'You have a duplicate term: {term.name}'
                    )
                term_names.add(term.name)
                if not term.logging and term.log_limit:
                    raise LimitButNoLogError(
                        f'Term {term.name}: Cannoy use log-limit without logging'
                    )
                if not chained_terms:
                    term.name = filter_name

                term = self.FixHighPorts(
                    term, af=filter_type, all_protocols_stateful=all_protocols_stateful
                )
                if not term:
                    continue

                new_terms.append(
                    self._TERM(
                        term,
                        filter_name,
                        all_protocols_stateful,
                        default_action,
                        filter_type,
                        self.verbose,
                        chained_terms,
                    )
                )

            self.iptables_policies.append(
                (header, filter_name, filter_type, default_action, new_terms)
            )

    def SetTarget(self, target: str, action: Optional[str] = None) -> None:
        """Sets policy's target and default action.

        Args:
          target: (string) target name
          action: (string) default action, only valid if target is a built-in chain
        """
        # there is only one item in iptables_policies
        pol = list(self.iptables_policies[0])
        pol[1] = target
        self._WarnIfCustomTarget(target)
        if action:
            pol[3] = action
        self.iptables_policies[0] = tuple(pol)

    def __str__(self) -> str:
        target = []
        pretty_platform = f'{self._PLATFORM[0].upper()}{self._PLATFORM[1:]}'

        if self._RENDER_PREFIX:
            target.append(self._RENDER_PREFIX)

        for header, filter_name, filter_type, default_action, terms in self.iptables_policies:
            # Add comments for this filter
            target.append(f'# {pretty_platform} {header.FilterName(self._PLATFORM)} Policy')

            # reformat long text comments, if needed
            comments = aclgenerator.WrapWords(header.comment, 70)
            if comments and comments[0]:
                for line in comments:
                    target.append(f'# {line}')
                target.append('#')
            # add the p4 tags
            target.extend(aclgenerator.AddRepositoryTags('# '))
            target.append(f"# {filter_type}")

            if filter_name in self._GOOD_FILTERS:
                if default_action:
                    target.append(self._DEFAULTACTION_FORMAT % (filter_name, default_action))
                elif self._PLATFORM == 'speedway':
                    # always specify the default filter states for speedway,
                    # if default action policy not specified for iptables, do nothing.
                    target.append(self._DEFAULTACTION_FORMAT % (filter_name, self._DEFAULT_ACTION))
            else:
                # Custom chains have no concept of default policy.
                target.append(self._DEFAULTACTION_FORMAT_CUSTOM_CHAIN % filter_name)
            # add the terms
            for term in terms:
                term_str = str(term)
                if term_str:
                    target.append(term_str)

        if self._RENDER_SUFFIX:
            target.append(self._RENDER_SUFFIX)

        target.append('')
        return '\n'.join(target)


class Error(aclgenerator.Error):
    """Base error class."""


class BadPortsError(Error):
    """Too many ports for a single iptables statement."""


class UnsupportedFilterError(Error):
    """Raised when we see an inappropriate filter."""


class NoIptablesPolicyError(Error):
    """Raised when a policy is received that doesn't support iptables."""


class TcpEstablishedError(Error):
    """Raised when a term has tcp-established option but not proto tcp only."""


class EstablishedError(Error):
    """Raised when a term has established option with inappropriate protocol."""


class UnsupportedDefaultActionError(Error):
    """Raised when a filter has an impermissible default action specified."""


class UnsupportedTargetOptionError(Error):
    """Raised when a filter has an impermissible default action specified."""


class LimitButNoLogError(Error):
    """Raised when log-limit is used by logging is not."""



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Demo generator for Aerleon."""

from absl import logging

from aerleon.lib import aclgenerator


class Term(aclgenerator.Term):
    """Used to create an individual term.

    The __str__ method must be implemented.

    Args: term policy.Term object

    This is created to be a demo.
    """

    _ACTIONS = {
        'accept': 'allow',
        'deny': 'discard',
        'reject': 'say go away to',
        'next': 'pass it onto the next term',
        'reject-with-tcp-rst': 'reset',
    }

    def __init__(self, term, term_type):
        self.term = term
        self.term_type = term_type

    def __str__(self):
        ret_str = []

        # NAME
        ret_str.append(f"{' ' * 4}Term: {self.term.name}{{")

        # COMMENTS
        if self.term.comment:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}#COMMENTS")
            for comment in self.term.comment:
                for line in comment.split('\n'):
                    ret_str.append(f"{' ' * 8}#{line}")

        # SOURCE ADDRESS
        source_address = self.term.GetAddressOfVersion(
            'source_address', self.AF_MAP.get(self.term_type)
        )
        source_address_exclude = self.term.GetAddressOfVersion(
            'source_address_exclude', self.AF_MAP.get(self.term_type)
        )
        if source_address:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}Source IP's")
            for saddr in source_address:
                ret_str.append(' ' * 8 + str(saddr))

        # SOURCE ADDRESS EXCLUDE
        if source_address_exclude:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}Excluded Source IP's")
            for ex in source_address:
                ret_str.append(' ' * 8 + str(ex))

        # SOURCE PORT
        if self.term.source_port:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}Source ports")
            ret_str.append(' ' * 8 + self._Group(self.term.source_port))

        # DESTINATION
        destination_address = self.term.GetAddressOfVersion(
            'destination_address', self.AF_MAP.get(self.term_type)
        )
        destination_address_exclude = self.term.GetAddressOfVersion(
            'destination_address_exclude', self.AF_MAP.get(self.term_type)
        )
        if destination_address:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}Destination IP's")
            for daddr in destination_address:
                ret_str.append(' ' * 8 + str(daddr))

        # DESINATION ADDRESS EXCLUDE
        if destination_address_exclude:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}Excluded Destination IP's")
            for ex in destination_address_exclude:
                ret_str.append(' ' * 8 + str(ex))

        # DESTINATION PORT
        if self.term.destination_port:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}Destination Ports")
            ret_str.append(' ' * 8 + self._Group(self.term.destination_port))

        # PROTOCOL
        if self.term.protocol:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}Protocol")
            ret_str.append(' ' * 8 + self._Group(self.term.protocol))

        # OPTION
        if self.term.option:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}Options")
            for option in self.term.option:
                ret_str.append(' ' * 8 + option)

        # ACTION
        for action in self.term.action:
            ret_str.append(' ')
            ret_str.append(f"{' ' * 8}Action: {self._ACTIONS.get(str(action))} all traffic")
        return '\n '.join(ret_str)

    def _Group(self, group):
        def _FormattedGroup(el):
            if isinstance(el, str):
                return el.lower()
            elif isinstance(el, int):
                return str(el)
            elif el[0] == el[1]:
                return '%d' % el[0]
            else:
                return '%d-%d' % (el[0], el[1])

        if len(group) > 1:
            rval = ''
            for item in group:
                rval = f"{rval}{item[0]!s} "
        else:
            rval = _FormattedGroup(group[0])
        return rval


class Demo(aclgenerator.ACLGenerator):
    """Demo rendering class.

    This class takes a policy object and renders output into
    a syntax which is not useable by routers. This class should
    only be used for testing and understanding how to create a
    generator of your own.

    Args:
      pol: policy.Policy object
    Steps to implement this library
    1) Import library in aclgen.py
    2) Create a 3 letter entry in the table in the render_filters
         function for the demo library and set it to False
    3) In the for header in policy.headers: use the previous entry
         to add an if statement to create a deep copy of the
         policy object
    4) Create an if statement that will be used if that specific
         policy object is present will pass the policy file
         onto the demo Class.
    5) The returned object can be then printed to a file using the
         do_output_filter function
    6) Create a policy file with a target set to use demo
    """

    _PLATFORM = 'demo'
    _SUFFIX = '.demo'

    _OPTIONAL_SUPPORTED_KEYWORDS = {
        'expiration',
    }

    def _TranslatePolicy(self, pol, exp_info):
        self.demo_policies = []
        for header, terms in pol.filters:
            filter_options = header.FilterOptions('demo')
            filter_name = filter_options[0]
            if len(filter_options) > 1:
                interface_specific = filter_options[1]
            else:
                interface_specific = 'none'
            filter_type = 'inet'
            term_names = set()
            new_terms = []
            for term in terms:
                if term.name in term_names:
                    raise DemoFilterError('Duplicate term name')
                term_names.add(term.name)
                new_terms.append(Term(term, filter_type))
            self.demo_policies.append(
                (header, filter_name, filter_type, interface_specific, new_terms)
            )

    def __str__(self):
        target = []
        for header, filter_name, filter_type, interface_specific, terms in self.demo_policies:
            target.append('Header {')
            target.append(' ' * 4 + 'Name: %s {' % filter_name)
            target.append(f"{' ' * 8}Type: {filter_type} ")
            for comment in header.comment:
                for line in comment.split('\n'):
                    target.append(f"{' ' * 8}Comment: {line}")
            target.append(f"{' ' * 8}Family type: {interface_specific}")
            target.append(f"{' ' * 4}}}")
            for term in terms:
                target.append(str(term))
                target.append(f"{' ' * 4}}}")
                target.append(' ')
            target.append('}')
        return '\n'.join(target)


class Error(Exception):
    pass


class DemoFilterError(Error):
    pass



# Copyright 2020-2021 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
"""A module to handle merging file configurations with CLI configs for Aerleon."""

import pathlib
from typing import Optional

import yaml

defaults = {
    'base_directory': './policies',
    'definitions_directory': './def',
    'policy_file': None,
    'output_directory': './',
    'optimize': False,
    'recursive': True,
    'debug': False,
    'verbose': False,
    'ignore_directories': ['DEPRECATED', 'def'],
    'max_renderers': 10,
    'shade_check': False,
    'exp_info': 2,
}

DEFAULT_FILE = './aerleon.yml'


class ConfigFileError(Exception):
    """Raised if there is some problem reading a config file."""


def load_config(
    config_file: Optional["str | pathlib.Path | list[str | pathlib.Path]"] = None,
    apply_defaults: bool = True,
) -> dict:
    """Load Aerleon configuration file(s).

    Args:
        config_file: An optional string or pathlib.Path with the location of the config file
            to open. A list of config files can be given. Default is './aerleon.yml'.
        apply_defaults: Whether to set missing config fields to their default values. Default
            is True.

    Raises:
        ConfigFileError: If the config file could not be opened and loaded. It is not an error
            if config_file is not given and the default config file is not found.

    Returns: A dictionary containing the contents of the config file. If no config file is found
        all default values will be returned (unless apply_defaults is False). If a list of config
        files is given their contents will be merged by their order in the list."""

    using_default = False
    if not config_file:
        config_file = [pathlib.Path(DEFAULT_FILE)]
        using_default = True
    elif not isinstance(config_file, list):
        config_file = [config_file]

    local_defaults = {}
    if apply_defaults:
        local_defaults.update(defaults)

    for config in config_file:
        if isinstance(config, str):
            config = pathlib.Path(config)

        try:
            with open(config) as f:
                data = yaml.safe_load(f)

                if not data or not isinstance(data, dict):
                    raise ConfigFileError(f"Config file contents not valid: {config}")

                local_defaults.update(data)  # NOTE: Safe as long as the config file is flat

        # If a user-specified config file is not found, re-raise
        # It is not an error if there is no default config file
        except FileNotFoundError as e:
            if not using_default:
                raise ConfigFileError(f"Config file not found: {config}") from e

        except PermissionError as e:
            raise ConfigFileError(f"Insufficient permissions to open config file: {config}") from e

        except IsADirectoryError as e:
            raise ConfigFileError(f"Expected a config file, found a directory: {config}") from e

        except OSError as e:
            raise ConfigFileError(f"Unable to open config file: {config}") from e

        except yaml.YAMLError as e:
            raise ConfigFileError(f"Config file is not a valid YAML file: {config}") from e

    return local_defaults
